import unittest
from unittest.mock import patch
import urllib.parse
import os
import re
import requests
from eduvpn.server import CustomServer
from .utils import (
    remove_existing_config, create_test_app,
    skip_if_network_manager_not_supported,
)
from .state_utils import StateTestCaseMixin


TEST_SERVER_ENV_VAR = 'TEST_SERVER'


def replace_url_path(url, path):
    parsed = urllib.parse.urlparse(url)
    parsed = parsed._replace(path=path, query=None)
    return urllib.parse.urlunparse(parsed)


def get_form_action(html):
    action_match = re.search(r'<form ([^>/]* )?action="(?P<action>[^"]+)"', html)
    if action_match is None:
        return action_match
    else:
        return action_match.group('action')


class ConnectTests(StateTestCaseMixin, unittest.TestCase):
    @skip_if_network_manager_not_supported
    @unittest.skipUnless(
        TEST_SERVER_ENV_VAR in os.environ,
        f"No test server given in environment variable {TEST_SERVER_ENV_VAR}",
    )
    def test_connect(self):
        from eduvpn.interface import state as interface_state
        from eduvpn import network as network_state

        if not os.environ[TEST_SERVER_ENV_VAR]:
            self.fail(f'empty value for {TEST_SERVER_ENV_VAR}')
        test_server = re.match(
            r'(?P<username>.+):(?P<password>.+)@(?P<address>.+)',
            os.environ[TEST_SERVER_ENV_VAR])
        if not test_server:
            self.fail(f'invalid value for {TEST_SERVER_ENV_VAR}')

        remove_existing_config()
        with create_test_app() as app:
            self.assertReachesInterfaceState(app, interface_state.ConfigurePredefinedServer)
            self.assertIsNone(app.interface_state.results)
            self.assertReachesNetworkState(app, network_state.UnconnectedState)

            # enter the server address
            app.interface_transition('enter_custom_address', test_server['address'])
            self.assertReachesInterfaceState(app, interface_state.ConfigureCustomServer)
            self.assertEqual(app.interface_state.address, test_server['address'])

            # perform the oauth login
            server = CustomServer(test_server['address'])
            with patch('webbrowser.open') as webbrowser_open:
                app.interface_transition('connect_to_server', server)
                self.assertReachesInterfaceState(app, interface_state.OAuthSetup)
                webbrowser_open.assert_called_once()
                url = webbrowser_open.call_args[0][0]

            # get the login form so we know where to post the data
            session = requests.Session()
            response = session.get(url)
            self.assertIn('<h1>Sign In</h1>', response.text)
            self.assertIn('name="userName"', response.text)
            self.assertIn('name="userPass"', response.text)
            self.assertIn('name="_form_auth_redirect_to"', response.text)
            action = get_form_action(response.text)
            self.assertIsNotNone(action)
            self.assertTrue(action.startswith('/'))
            auth_url = replace_url_path(response.url, action)

            # post the login credentials
            response = session.post(auth_url, data=dict(
                userName=test_server['username'],
                userPass=test_server['password'],
                _form_auth_redirect_to=url,
            ))
            if 'The credentials you provided were not correct.' in response.text:
                self.fail("invalid credentials for test server")
            self.assertIn('<h1>Approve Application</h1>', response.text)

            # post the approval for the eduvpn app
            response = session.post(response.url, data=dict(approve='yes'))
            self.assertIn('You can now close this window', response.text)

            # now wait for the connection to be established
            self.assertReachesInterfaceState(app, interface_state.ConnectionStatus)
            self.assertReachesNetworkState(app, network_state.ConnectedState)

            app.network_transition('disconnect')
            self.assertReachesNetworkState(app, network_state.DisconnectedState)



from datetime import datetime, timedelta, timezone
from . import settings
from .state_machine import BaseState
from .app import Application
from .server import ConfiguredServer as Server


def now() -> datetime:
    return datetime.now(timezone.utc)


class Validity:
    def __init__(self, start: datetime, end: datetime):
        assert start.tzinfo is not None
        assert end.tzinfo is not None
        self.start = start
        self.end = end

    @property
    def duration(self) -> timedelta:
        """
        Return the duration of this validity.
        """
        return self.end - self.start

    @property
    def remaining(self) -> timedelta:
        """
        Return the duration from now until expiry.
        """
        return self.end - now()

    @property
    def pending_remaining(self) -> timedelta:
        """
        Return the duration from now until pending expiry.
        """
        return self.pending_expiry_time - now()

    def fraction(self, fraction: float) -> datetime:
        return self.start + self.duration * fraction

    @property
    def pending_expiry_time(self) -> datetime:
        """
        Determine the moment when the user should be notified
        of the sessions expiry.
        """
        return max(
            self.fraction(settings.SESSION_PENDING_EXPIRY_FRACTION),
            self.end - timedelta(minutes=settings.SESSION_PENDING_EXPIRY_MINUTES)
        )

    @property
    def is_pending_expiry(self) -> bool:
        """
        Return True if the validity is pending expiry or has already expired.
        """
        return now() >= self.pending_expiry_time

    @property
    def is_expired(self) -> bool:
        """
        Return True if the validity has expired.
        """
        return now() >= self.end


class SessionState(BaseState):
    """
    Base class for all session states.
    """

    def new_session(self, app: Application, server: Server, validity: Validity) -> 'SessionState':
        if validity.is_expired:
            return SessionExpiredState(server, validity)
        elif validity.is_pending_expiry:
            return SessionPendingExpiryState(server, validity)
        else:
            return SessionActiveState(server, validity)


class InitialSessionState(SessionState):
    """
    The state of the session when the app starts.

    This is a transient state until
    the actual session state is obtained.
    """

    is_active = False

    def found_active_session(self, app: Application, server: Server, validity: Validity) -> SessionState:
        """
        An already active session was found.
        """
        app.network_transition('found_previous_connection')
        return self.new_session(app, server, validity)

    def no_previous_session_found(self, app: Application) -> SessionState:
        """
        No previously established session was found.
        """
        app.network_transition('no_previous_connection_found')
        return NoSessionState()


class NoSessionState(SessionState):
    """
    No session is currently active.
    """

    is_active = False


class SessionActiveState(SessionState):
    """
    A session is currently active.
    """

    is_active = True

    def __init__(self, server: Server, validity: Validity):
        self.server = server
        self.validity = validity

    def pending_expiry(self, app: Application) -> SessionState:
        return SessionPendingExpiryState(self.server, self.validity)


class SessionPendingExpiryState(SessionState):
    """
    The session is about to expire.
    """

    is_active = True

    def __init__(self, server: Server, validity: Validity):
        self.server = server
        self.validity = validity

    def has_expired(self, app: Application) -> SessionState:
        return SessionExpiredState(self.server, self.validity)

    def renew(self, app: Application) -> SessionState:
        app.interface_transition('renew_session', self.server)
        return self


active_session_states = (
    SessionActiveState,
    SessionPendingExpiryState,
)


class SessionExpiredState(SessionState):
    """
    The session has expired.
    """

    is_active = False

    def __init__(self, server: Server, validity: Validity):
        self.server = server
        self.validity = validity

    def renew(self, app: Application) -> SessionState:
        app.interface_transition('renew_session', self.server)
        return self



from contextlib import contextmanager
import threading
from shutil import rmtree
import unittest
from eduvpn import settings
import eduvpn.nm
from eduvpn.variants import EDUVPN
from eduvpn.utils import run_in_main_gtk_thread
from eduvpn.app import Application


def remove_existing_config():
    rmtree(settings.CONFIG_PREFIX, ignore_errors=True)


@contextmanager
def create_test_app():
    from gi.repository import GLib
    event_loop = GLib.MainLoop()
    thread = threading.Thread(target=event_loop.run)
    thread.start()
    variant = EDUVPN
    app = Application(variant, make_func_threadsafe=run_in_main_gtk_thread)
    app.initialize()
    try:
        yield app
    finally:
        event_loop.quit()


platform_supports_network_manager = eduvpn.nm.NM is not None

skip_if_network_manager_not_supported = unittest.skipUnless(
    platform_supports_network_manager,
    "Network Manager not supported",
)



mock_config = """# OpenVPN Client Configuration
dev tun
client
nobind
remote-cert-tls server
comp-lzo
verb 3
server-poll-timeout 10
auth SHA256
cipher AES-256-CBC
tls-version-min 1.2
tls-cipher TLS-ECDHE-RSA-WITH-AES-256-GCM-SHA384
<ca>
-----BEGIN CERTIFICATE-----
MIIFJDCCAwygAwIBAgIJAJ1NwjmG+n/3MA0GCSqGSIb3DQEBCwUAMBExDzANBgNV
BgNVBAMTBlZQTiBDQTCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBALOW
Zak2nNf7uZ1+XaoR78gkBHqszXlaFdPByC2bQn0vrsGurUHTXO843MDfsckbeNEE
8ZrqXnNnqaojP90UkGIXbXLl7U8kborvKF/gQm6gsk3O0obhTVDdEj3kYKFGxmqu
QHJo3C2amro4wVIv/BhXzIv6oP11Va/6FcwAYj8disNcxJgXWSUdARrEIC3PME58
Ld+a2jvXL/m8xBp+qVThvfmq2Pak1olhbgDueUSsQYiQZMq2rX3/1f+SInxZt3kJ
BAMTBlZQTiBDQTAeFw0xNzAyMjcxNDA5NTJaFw0yMjAyMjcxNDA5NTJaMBExDzAN
0LWOOSBzGwYnjfGmm/JzM06jHq+vyfrFwl28uNuCVRwJpvmjJC8+O8TYFgOrx8Oe
Q7gbC+Wrx1Nv/u2j59ofTPk2R5oNdrYt3h5YVv0WocHmiYb+CzdwxBvyZ50Uy1+x
0XEZpfYV4x6m1/8sEqYANx+5U+3KXZznUXlevyLqgS5tn3hGuxwRuxb5AifwRRmF
z5wSVuDilCGnzNGyqJvP4rKAlGakQx8nyFIsdMsa7OpQVd4n+ex5HhfMlSa7Bvpj
DK+8p3ynd11jihCkFAzOqldBN2hR1gCjx2b52zLk0vzTyyiCEebqb9hqKf+jb3/j
FyE9gzlm2uR9rvvyVKnsDnuCb5vsnWdPTUvBov6tAgMBAAGjfzB9MB0GA1UdDgQW
BBTmp4VYXwUkBWl7cK4WspgpaIyI8zBBBgNVHSMEOjA4gBTmp4VYXwUkBWl7cK4W
8CKV1gYtjPBc6OOMNYyXfXlu0qtkLljV/Pdt4hxUNyyptPvJgP8xLg+z+RoZZmxV
spgpaIyI86EVpBMwETEPMA0GA1UEAxMGVlBOIENBggkAnU3COYb6f/cwDAYDVR0T
bifEWHPpbS1RceRBzk+QUNjRmQrT+hCdzpHro8SdbaPg7eRG7HvC5zi6tvxLirn/
0ThsnRb5xWkDYQ2BXponBTnFRyYaiaXyEuJTAvfbF51R3ycy03tl4v15UtzW/Spa
n1CqNVHH+vinqSo1cXV17b9PFcfohA2n86tDwkqOC+0fPX3KO2v8ALANxNYEunzW
xWPQM0wTgoXuUpOx8/UkGYVxOnwy9uHtL4HS1+5tyTjKv5Ld8NktDPiNvSHTV8zt
BAUwAwEB/zALBgNVHQ8EBAMCAQYwDQYJKoZIhvcNAQELBQADggIBACIReoZBwxA+
S5sb3PokhBktJxEy8MzamAwuJEpiLikYPUJvoNI1XzH9JJR6cI1rfGtABb8GCikA
9KxDmz8SBuJVjYuWA/d0DlLAAhUiq+gIkuh1YaC6dhjUxhQAxMe/WW38KHV9BlrR
WrnstYLLSuVkwLmg5h+ltra2rqqyRTYX66+eKF8gj4p/xP0JAiAsBSLKztuSodce
rhYC3Jy9EQ0MBP0oujwdOLSLMoGqM4xZPiMmsELgOVoeX5ypk2DPpWwNeOYVsDrV
b2VJ/qPt0tfQjGIEYba34wNAsmuI/WLjayKVHz6zwG1nueG+9/JHVW1lYOv3iwXe
PRTJkpeEl44pu4vsvU/XyAoUijFwwuazzbqLTNwc7h7kM7Vq/beCyN3HkyW44sla
nIYUto+cPuPcEEuvyVeOT6UjzKgnLbmY
-----END CERTIFICATE-----
</ca>
key-direction 1
<tls-auth>
#
# 2048 bit OpenVPN static key
#
-----BEGIN OpenVPN Static key V1-----
bbb4a1bc4e97ce6cfa7ea930c3bb7feb
e1af60b6b06a6acd39c38c29d6c0197c
85dab27a52c30d0e31b4ce80bfb61dab
69577c09b349787b7e9336a7e655a5bc
51c3ef467fdf867a1cc951ae6ebd995a
ba096690d73fbd8cd41ffff231c1c17e
ef022cedf84639dc04b8fb204066f31e
484804c5535d50c428794a68ba65b432
781c4e76285ba799cbd81bb4473b3f45
d7b127ab548377645dd82c14fd5db3b2
8487de31db848a1a476616ab4650a55c
e1a42d4a8521218ac56789e3c10848e7
8349bea907fcbc333282879306cee469
b0b85a410294cdea5eb29ae009f98687
8cb2f25feff977b15b3249bee779659c
367e7c0aa4d23d8af9d948f992fbd10a
-----END OpenVPN Static key V1-----
</tls-auth>
remote internet.demo.eduvpn.nl 1194 udp
remote internet.demo.eduvpn.nl 1194 tcp
remote internet.demo.eduvpn.nl 443 tcp
"""

mock_cert = """
-----BEGIN CERTIFICATE-----
MIIFSzCCAzOgAwIBAgICAdkwDQYJKoZIhvcNAQELBQAwETEPMA0GA1UEAxMGVlBO
IENBMB4XDTE3MDgyNDE0MDk0NloXDTE4MDgyNDE0MDk0NlowKzEpMCcGA1UEAxMg
ZDU1MDhmMjk3MjRmZTc4MTEyZTViZGY1MmZjOWUwMTYwggIiMA0GCSqGSIb3DQEB
AQUAA4ICDwAwggIKAoICAQCZ5/FNshijQG51CVpyXdYpk6gnQbEX9ZKTW5S2HULB
Z0jmVOisdW6eV8nHTCkpZx8MWRvIe7OvndBms4I8/31s30sQM3UzLpQqUelS6DvG
hNT5I9AYzrISOFkuZ0zUIRQYTkHNEBzYxkEqYqMvOMecY48CQ5CX6PY8TI390ge4
kzdllrHjr8c8LjnmHa3uHwruRrzFFtnQFHqCbr9hBaDEK8ShVLmHdHvpY43aGEec
bqCzPwemp8+NW9Ibv+iocj4ISJfMKSHAhV8PogIYT0mNyp1O3OMV8XEib1/lEj7r
EJsAcnnNjL/iY6fqNNcxHC0az0GPkuymiHEh5p17/wgBrHgtN6ukAIskUbhcSh7c
63qy5n3xJvpu6x9YF57rXl6eRG9ybVYu4SjJRVQnBcK56KTrJ9kJk3j7W5YWxOSf
fJ1KyFQtp0XKS4QndE5rvu+pEXvUAGJFZ31N1ZmAQzrUINUNLz89vuuBYE5Nxtk5
HQFmEqZ934EtvcHRl719fl0Dd7QnT6f2Ph9/TxX8K7yye1wKwFcPlTVoJs+8/Rlo
ssjbwMDY+dpSrtMRSj9VaEkJNKqNhKFAOshvaru3eeWjr4yBu762r1Npw0iczRED
9DLJM7j1zsEEGVYyIqA2MAK91FYraMT1jjKG1mox1Vm+CElWY72coNN7+a+e7f/5
/dMl6s08MEEGA1UdIwQ6MDiAFOanhVhfBSQFaXtwrhaymClojIjzoRWkEzARMQ8w
0wIDAQABo4GSMIGPMAkGA1UdEwQCMAAwHQYDVR0OBBYEFOHZWTq3OVB886F023oQ
BgNVHQ8EBAMCB4AwDQYJKoZIhvcNAQELBQADggIBAKEajJgeStmXI7B2KeSNDBVt
DQYDVQQDEwZWUE4gQ0GCCQCdTcI5hvp/9zATBgNVHSUEDDAKBggrBgEFBQcDAjAL
NYm1JwJwegYk6PeiDIvLYl4FrUznNYHT3udOd9ikn/vFRCEexPTksasNpka+tRme
PH82cZtwHZZgajpUa640vQXW7nmZC6YdTkZW74he1Zqi28FNMNAQ7dzrDzMdVOyv
NSbQQTX0Qf6Yobocb8jmHivonAR7dZ1MzGPyf6w7OqkQ+7mszOX0o2IZFsAws3L0
PCzqUP9LwZCk+gtA5mmaaXXqMMhoL9/QjFNJBtxllz9ksMyCnfQDRTZECjNsODH4
JcBcJPTWQ1c67XJNDXaJ3QMRlJRqdHAGJxg1Si4ajsoHD6fVqpKvCwyhIFSGMQhW
eN4i+8OTrr94YcLpBrP4oK1zTYfUdDDk/SP6Ax+xJEIYbLmJR5jci9di0Kdc4FJL
fMmH62tFb/U7nYEwT3V9+IXFD6eljaK8k2E1VISzPV/mT5RkmBGzx/MB4NHUX35U
T93kKrIIM0/VXzaMOr2iVTuc90im5yFczFXuy0JhL1agV7yCEW43YUClGESg48W6
3rp1ufZWraHITnyUPcHObgq8/51uKHkPIpNx1F4rqrXlrgZJfjANVO3MLJkxvdRA
zxWzRatGON7I9fPr4zd9h6au5rN9iOObof+JZGPk9tbH2Bg6wV3qZNIPZPVUBfc4
q5QpoF0ATpXqjhwu4yI9
-----END CERTIFICATE-----
"""
mock_key = """
-----BEGIN PRIVATE KEY-----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-----END PRIVATE KEY-----
"""

mock_config_dict = {
    'auth': 'SHA256',
    'ca': '\n-----BEGIN CERTIFICATE-----\nMIIFJDCCAwygAwIBAgIJAJ1NwjmG+n/3MA0GCSqGSIb3DQEBCwUAMBExDzANBgNV\nBgNVBAMTBlZQTiBDQTCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIBALOW\nZak2nNf7uZ1+XaoR78gkBHqszXlaFdPByC2bQn0vrsGurUHTXO843MDfsckbeNEE\n8ZrqXnNnqaojP90UkGIXbXLl7U8kborvKF/gQm6gsk3O0obhTVDdEj3kYKFGxmqu\nQHJo3C2amro4wVIv/BhXzIv6oP11Va/6FcwAYj8disNcxJgXWSUdARrEIC3PME58\nLd+a2jvXL/m8xBp+qVThvfmq2Pak1olhbgDueUSsQYiQZMq2rX3/1f+SInxZt3kJ\nBAMTBlZQTiBDQTAeFw0xNzAyMjcxNDA5NTJaFw0yMjAyMjcxNDA5NTJaMBExDzAN\n0LWOOSBzGwYnjfGmm/JzM06jHq+vyfrFwl28uNuCVRwJpvmjJC8+O8TYFgOrx8Oe\nQ7gbC+Wrx1Nv/u2j59ofTPk2R5oNdrYt3h5YVv0WocHmiYb+CzdwxBvyZ50Uy1+x\n0XEZpfYV4x6m1/8sEqYANx+5U+3KXZznUXlevyLqgS5tn3hGuxwRuxb5AifwRRmF\nz5wSVuDilCGnzNGyqJvP4rKAlGakQx8nyFIsdMsa7OpQVd4n+ex5HhfMlSa7Bvpj\nDK+8p3ynd11jihCkFAzOqldBN2hR1gCjx2b52zLk0vzTyyiCEebqb9hqKf+jb3/j\nFyE9gzlm2uR9rvvyVKnsDnuCb5vsnWdPTUvBov6tAgMBAAGjfzB9MB0GA1UdDgQW\nBBTmp4VYXwUkBWl7cK4WspgpaIyI8zBBBgNVHSMEOjA4gBTmp4VYXwUkBWl7cK4W\n8CKV1gYtjPBc6OOMNYyXfXlu0qtkLljV/Pdt4hxUNyyptPvJgP8xLg+z+RoZZmxV\nspgpaIyI86EVpBMwETEPMA0GA1UEAxMGVlBOIENBggkAnU3COYb6f/cwDAYDVR0T\nbifEWHPpbS1RceRBzk+QUNjRmQrT+hCdzpHro8SdbaPg7eRG7HvC5zi6tvxLirn/\n0ThsnRb5xWkDYQ2BXponBTnFRyYaiaXyEuJTAvfbF51R3ycy03tl4v15UtzW/Spa\nn1CqNVHH+vinqSo1cXV17b9PFcfohA2n86tDwkqOC+0fPX3KO2v8ALANxNYEunzW\nxWPQM0wTgoXuUpOx8/UkGYVxOnwy9uHtL4HS1+5tyTjKv5Ld8NktDPiNvSHTV8zt\nBAUwAwEB/zALBgNVHQ8EBAMCAQYwDQYJKoZIhvcNAQELBQADggIBACIReoZBwxA+\nS5sb3PokhBktJxEy8MzamAwuJEpiLikYPUJvoNI1XzH9JJR6cI1rfGtABb8GCikA\n9KxDmz8SBuJVjYuWA/d0DlLAAhUiq+gIkuh1YaC6dhjUxhQAxMe/WW38KHV9BlrR\nWrnstYLLSuVkwLmg5h+ltra2rqqyRTYX66+eKF8gj4p/xP0JAiAsBSLKztuSodce\nrhYC3Jy9EQ0MBP0oujwdOLSLMoGqM4xZPiMmsELgOVoeX5ypk2DPpWwNeOYVsDrV\nb2VJ/qPt0tfQjGIEYba34wNAsmuI/WLjayKVHz6zwG1nueG+9/JHVW1lYOv3iwXe\nPRTJkpeEl44pu4vsvU/XyAoUijFwwuazzbqLTNwc7h7kM7Vq/beCyN3HkyW44sla\nnIYUto+cPuPcEEuvyVeOT6UjzKgnLbmY\n-----END CERTIFICATE-----\n',
    'cert': '\n-----BEGIN CERTIFICATE-----\nMIIFSjCCAzKgAwIBAgIBXzANBgkqhkiG9w0BAQsFADARMQ8wDQYDVQQDDAZWUE4g\nQ0EwHhcNMTgwNzIwMTQwMTUxWhcNMTkwMTE2MTQwMTUxWjArMSkwJwYDVQQDDCA5\nZjQzOTUzZjYzNzEyMTIxMzBkMmY4ZDY1YmFkODY5NDCCAiIwDQYJKoZIhvcNAQEB\nBQADggIPADCCAgoCggIBAMlxDDud9y8mxKWQjxkXUWcyaVdWEbw8GGdkZJA5SAgM\nXvQ2uIggKgQ7fZoptsRyWBeJQFi0F4IuwGICaJhLwPgCtd5okt2WpQEvDiBKEYqZ\nXY/LfiTEclRKJG9yI7r/Z8m85Hlm/V/EbDZrWPhzk3smh1KlAYY3saJr+HsoaAcv\nRCWA3mWefux/d0tsP3EOuQu4zV+Oak75j7Kdt7AygOk29SG5SYDqSGAkOf2iO8Ez\nN3PCx/zEjWy2CAacdoIpZpxWbZM+01YIkgFLIn5h+qo8DEEkeIfB6k/gKMDd8bvF\nkeE87q5kWFTO+U9BigyPc31opNILT5Mwy1xBXt0h3tSksXOKOfwTQgpnpPABaSFN\n0Tdl7A918CfiU6Mhhrm+wT45nx+BWzl3yAEpO57oe4kJaODb0/YVbnv5aZ7Mq3Sj\nxoGhBESaY15dAWaupq8gduwhJojdpUnMwlNCvJtOSn1NxRWdBRELtN3irUxhO4Wq\nYi8esnzgCUMeBf+h2Y2hazCw1HrNTLjyHAtkJsLw9DR7s3V4juHPgWfYTcUm/NeU\nXW+EvauxxDZWUwcd3S1PjGgCArFDHRqgXnSsd+NJYeygDcNKluQGcHJokUDtSEI2\nGWRbfeueysyzqeO0mhLsLL/BLNlb7fEqEMXeZrUJ15VbCW/rckrvuwAc2pE9wOaH\nAgMBAAGjgZIwgY8wCQYDVR0TBAIwADAdBgNVHQ4EFgQUZoAk4rVSPmOlz53LUZFd\n8cmso30wQQYDVR0jBDowOIAUDdOUbUGNeDKmECUa1UNkuIPB3GWhFaQTMBExDzAN\nBgNVBAMMBlZQTiBDQYIJAIo819EOTUu+MBMGA1UdJQQMMAoGCCsGAQUFBwMCMAsG\nA1UdDwQEAwIHgDANBgkqhkiG9w0BAQsFAAOCAgEAI3pD/f7MS6NWcpJ/t4qb/aiv\n4m2rRxOxOewVbzV0lPAaazJ70kUMdVGHsiaBQsLYgTQ2korsvKuHDZ/cH6bn0WYx\nF5ZvPrTktAFS2EiFyaMYbyZT6hGCJQaUbhYV7UXD1om1epo5dPPCJrXau2AGa8Y2\n9fPXXbilUeR/Rh1xiHWDn0Vfg9hnPnHvs8Vt1OfdpCRLD27S1DL4Svj9Vv4prya+\n6MxYVi1obc7oEO8y9mVb6Wq7JyOqfldWjUOxyY9tzl5peK8HcyrPmv/LLr5DAdVe\nmACqmXgwX/L+87X+S+RBy+71yd2zZhQzg8M1NDZJbxpCzhNylhQc1FLyUnO+p8ee\neT0IjR7SZuVYaTsd+HjI3TnBXNztNHSwUmMXBsI/1qrDXC7oqPBS7bDFc65EqIXt\n7PnP2ta1posklLq1Df5vyWqCxX3yu2o90Ph+Y4yoDV6PSPmtWMzDbmfh8RzjvD7b\nyCHzrmALmkZIXTo9XLYRET1zAhRwH0zrHXhp/5ae7lD0x/f0sIJnKfpv7kRIbVhO\neQ1X9t85MOvUuVVvHh1s6aUkB3/XNso3JHmg0YQK/X19AYjU6CIgYuDFa8YCwRfP\nJo/GJNGA/Ga8EnF2dEQVPtG0haONFcsUH/CWvGxdisfEPgYgyXSPH+qLdMENWzBN\nT6AhwlmSLX1GFPpwQeI=\n-----END CERTIFICATE-----\n',
    'cipher': 'AES-256-CBC',
    'client': None,
    'comp-lzo': None,
    'dev': 'tun',
    'key': '\n-----BEGIN PRIVATE KEY-----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\nxL2Pccm97PHvBS/J8JMzBn8y3hhT0g==\n-----END PRIVATE KEY-----\n',
    'key-direction': '1',
    'nobind': None,
    'remote': [['internet.demo.eduvpn.nl', '1194', 'udp'],
               ['internet.demo.eduvpn.nl', '1194', 'tcp'],
               ['internet.demo.eduvpn.nl', '443', 'tcp']],
    'remote-cert-tls': 'server',
    'server-poll-timeout': '10',
    'tls-auth': '\n#\n# 2048 bit OpenVPN static key\n#\n-----BEGIN OpenVPN Static key V1-----\nbbb4a1bc4e97ce6cfa7ea930c3bb7feb\ne1af60b6b06a6acd39c38c29d6c0197c\n85dab27a52c30d0e31b4ce80bfb61dab\n69577c09b349787b7e9336a7e655a5bc\n51c3ef467fdf867a1cc951ae6ebd995a\nba096690d73fbd8cd41ffff231c1c17e\nef022cedf84639dc04b8fb204066f31e\n484804c5535d50c428794a68ba65b432\n781c4e76285ba799cbd81bb4473b3f45\nd7b127ab548377645dd82c14fd5db3b2\n8487de31db848a1a476616ab4650a55c\ne1a42d4a8521218ac56789e3c10848e7\n8349bea907fcbc333282879306cee469\nb0b85a410294cdea5eb29ae009f98687\n8cb2f25feff977b15b3249bee779659c\n367e7c0aa4d23d8af9d948f992fbd10a\n-----END OpenVPN Static key V1-----\n',
    'tls-cipher': 'TLS-ECDHE-RSA-WITH-AES-256-GCM-SHA384',
    'tls-version-min': '1.2',
    'verb': '3'
}

mock_server = {
    'server_type': 'secure_internet',
    'base_url': 'https://eduvpn.bogus/',
    'public_key_list': ['Xv3l24gbMX8NtTnFQbWO2fGKPwKuc6EbjQDv8qwXXXX'],
    'country_code': 'NL',
    'support_contact': ['mailto:bogus@bogus.nl']
}

mock_org = {
    'display_name': {'nl': 'bogus', 'en': 'bogus'},
    'org_id': 'http://idp.mock.bogus/adfs/services/trust',
    'secure_internet_home': 'https://idp.mock.bogus/'
}



from unittest import TestCase, skip
from unittest.mock import patch
from .utils import (
    remove_existing_config, create_test_app,
    skip_if_network_manager_not_supported,
)
from .state_utils import StateTestCaseMixin


PROFILE_NAME_1 = 'Test Profile A'
PROFILE_NAME_2 = 'Test Profile B'


class TestOAuthSession:
    def refresh_token(self, token_url):
        pass

    def get(self, url):
        return TestOAuthResponse()


class TestOAuthResponse:
    status_code = 200

    def json(self):
        return dict(
            profile_list={'data': [
                dict(
                    profile_id='t1',
                    display_name=PROFILE_NAME_1,
                    two_factor=False,
                ),
                dict(
                    profile_id='t2',
                    display_name=PROFILE_NAME_2,
                    two_factor=False,
                ),
            ]},
        )


class FlowTests(StateTestCaseMixin, TestCase):
    @skip("disabled until someone mocks the network calls")
    @skip_if_network_manager_not_supported
    def test_first_start(self):
        from eduvpn.interface import state as interface_state
        from eduvpn import network as network_state

        remove_existing_config()
        with create_test_app() as app:
            self.assertReachesInterfaceState(app, interface_state.ConfigurePredefinedServer)
            self.assertIsNone(app.interface_state.results)
            self.assertReachesNetworkState(app, network_state.UnconnectedState)

            # search for a server called 'demo'
            app.interface_transition('enter_search_query', 'demo')
            self.assertReachesInterfaceState(app, interface_state.ConfigurePredefinedServer)
            self.assertEqual(
                list(map(str, app.interface_state.results)),
                ['Demo'],
            )
            server = app.interface_state.results[0]

            # perform the oauth login
            with patch('eduvpn.oauth2.run_challenge_in_background') as oauth_func:
                with patch('webbrowser.open') as webbrowser_open:
                    url = 'test-url'
                    webserver = object()
                    callback = None

                    def oauth_challenge(token_endpoint, auth_endpoint, app_variant, cb):
                        nonlocal callback
                        callback = cb
                        return webserver, url

                    oauth_func.side_effect = oauth_challenge
                    app.interface_transition('connect_to_server', server)
                    self.assertReachesInterfaceState(app, interface_state.OAuthSetup)
                    self.assertIs(app.interface_state.oauth_web_server, webserver)
                    webbrowser_open.assert_called_once_with(url)

            self.assertIsNotNone(callback)
            oauth_session = TestOAuthSession()
            callback(oauth_session)
            self.assertReachesInterfaceState(app, interface_state.ChooseProfile)

            self.assertEqual(
                list(map(str, app.interface_state.profiles)),
                [PROFILE_NAME_1, PROFILE_NAME_2],
            )



from datetime import datetime, timezone
from configparser import ConfigParser
from typing import Optional, List, Type
from . import nm
from .ovpn import Ovpn
from .server import Protocol
from .session import Validity
from .crypto import SecretKey
from .utils import parse_http_date_header, parse_http_expires_header


def get_response_date(response) -> datetime:
    try:
        return parse_http_date_header(response.headers['Date'])
    except (KeyError, ValueError):
        return datetime.now(timezone.utc)


class Connection:
    "Base class for connection configurations."

    protocol: Protocol
    validity: Validity

    def __init__(self, validity: Validity):
        self.validity = validity

    @classmethod
    def parse(cls, response) -> 'Connection':
        protocol = Protocol.get_by_config_type(response.headers['Content-Type'])
        connection_type = protocol_to_connection_type[protocol]
        return connection_type.parse(response)

    def set_secret_key(self, secret_key: SecretKey):
        pass

    def force_tcp(self):
        raise NotImplementedError

    def connect(self, callback):
        """
        Start this connection.

        This method is always called from the main thread.
        """
        raise NotImplementedError


class OpenVPNConnection(Connection):
    protocol = Protocol.OPENVPN

    def __init__(self, validity: Validity, ovpn: Ovpn):
        self.ovpn = ovpn
        super().__init__(validity)

    @classmethod
    def parse(cls, response) -> 'OpenVPNConnection':
        expiry = parse_http_expires_header(response.headers['Expires'])
        created = get_response_date(response)
        validity = Validity(created, expiry)
        ovpn = Ovpn.parse(response.text)
        return cls(validity=validity, ovpn=ovpn)

    def force_tcp(self):
        self.ovpn.force_tcp()

    def connect(self, callback):
        nm.start_openvpn_connection(
            self.ovpn,
            callback=callback,
        )


class WireGuardConnection(Connection):
    protocol = Protocol.WIREGUARD

    def __init__(self, validity: Validity, config: ConfigParser):
        self.config = config
        self.secret_key: Optional[SecretKey] = None
        super().__init__(validity)

    @classmethod
    def parse(cls, response) -> 'WireGuardConnection':
        expiry = parse_http_expires_header(response.headers['Expires'])
        created = get_response_date(response)
        validity = Validity(created, expiry)
        config = ConfigParser()
        config.read_string(response.text)
        return cls(validity=validity, config=config)

    def set_secret_key(self, secret_key: SecretKey):
        assert self.secret_key is None
        self.secret_key = secret_key

    def force_tcp(self):
        raise NotImplementedError("WireGuard cannot be forced over tcp")

    def connect(self, callback):
        assert self.secret_key is not None
        nm.start_wireguard_connection(
            self.config,
            secret_key=self.secret_key,
            callback=callback,
        )


connection_types: List[Type[Connection]] = [
    OpenVPNConnection,
    WireGuardConnection,
]

protocol_to_connection_type = {
    connection_type.protocol: connection_type
    for connection_type in connection_types
}



import webbrowser
from typing import Optional, Callable, Tuple
from requests_oauthlib import OAuth2Session
from ..settings import CLIENT_ID, SCOPE, CODE_CHALLENGE_METHOD
from ..variants import ApplicationVariant
from ..crypto import gen_code_verifier, gen_code_challenge
from .http import OAuthWebServer


class OAuthChallenge:
    def __init__(self, session: OAuth2Session, token_endpoint: str, verifier: str, url: str, state: str):
        self.session = session
        self.token_endpoint = token_endpoint
        self.verifier = verifier
        self.url = url
        self.state = state


def create_challenge(
    redirect_uri: str,
    token_endpoint: str,
    authorization_endpoint: str,
) -> OAuthChallenge:
    session = OAuth2Session(
        CLIENT_ID,
        redirect_uri=redirect_uri,
        auto_refresh_url=token_endpoint,
        scope=SCOPE,
    )
    code_verifier = gen_code_verifier()
    code_challenge = gen_code_challenge(code_verifier)
    authorization_url, state = session.authorization_url(
        url=authorization_endpoint,
        code_challenge_method=CODE_CHALLENGE_METHOD,
        code_challenge=code_challenge,
    )
    return OAuthChallenge(session, token_endpoint, code_verifier, authorization_url, state)


def setup_challenge(
    token_endpoint: str,
    authorization_endpoint: str,
    app_variant: ApplicationVariant,
) -> Tuple[OAuthChallenge, OAuthWebServer]:
    webserver = OAuthWebServer(app_variant)
    challenge = create_challenge(
        webserver.success_url,
        token_endpoint,
        authorization_endpoint,
    )
    return challenge, webserver


def perform_challenge(
    challenge: OAuthChallenge,
    webserver: OAuthWebServer,
) -> Optional[OAuth2Session]:
    data = webserver.run()
    if data is None:
        # The operation was cancelled by the user.
        return None
    fetch_token(challenge, data)
    return challenge.session


def run_challenge(
    token_endpoint: str,
    authorization_endpoint: str,
    app_variant: ApplicationVariant,
) -> Optional[OAuth2Session]:
    challenge, webserver = setup_challenge(
        token_endpoint, authorization_endpoint, app_variant)
    webbrowser.open(challenge.url)
    return perform_challenge(challenge, webserver)


def perform_challenge_in_background(
    challenge: OAuthChallenge,
    webserver: OAuthWebServer,
    callback: Callable[[Optional[OAuth2Session]], None],
):
    def complete_callback(data: dict):
        if data is None:
            # The operation was cancelled by the user.
            callback(None)
        else:
            fetch_token(challenge, data)
            callback(challenge.session)

    webserver.run_in_background(complete_callback)


def run_challenge_in_background(
    token_endpoint: str,
    authorization_endpoint: str,
    app_variant: ApplicationVariant,
    callback: Callable[[Optional[OAuth2Session]], None],
):
    challenge, webserver = setup_challenge(
        token_endpoint, authorization_endpoint, app_variant)
    perform_challenge_in_background(challenge, webserver, callback)
    return webserver, challenge.url


def fetch_token(challenge: OAuthChallenge, response: dict):
    assert challenge.state == response['state'][0]
    challenge.session.fetch_token(
        token_url=challenge.token_endpoint,
        code=response['code'][0],
        code_verifier=challenge.verifier,
        client_id=challenge.session.client_id,
        include_client_id=True,
    )



import gi
gi.require_version('Gtk', '3.0')  # noqa: E402
from gi.repository import Gtk


def show_ui_component(component, show: bool):
    """
    Set the visibility of a UI component.
    """
    if show:
        component.show()  # type: ignore
    else:
        component.hide()  # type: ignore


def link_markup(link: str) -> str:
    try:
        scheme, rest = link.split(':', 1)
        if rest.startswith('//'):
            rest = rest[2:]
    except ValueError:
        return link
    else:
        return f'<a href="{link}">{rest}</a>'


def show_error_dialog(parent, name: str, title: str, message: str):
    dialog = Gtk.MessageDialog(  # type: ignore
        parent=parent,
        type=Gtk.MessageType.INFO,  # type: ignore
        buttons=Gtk.ButtonsType.OK,  # type: ignore
        title=name,
        message_format=title)
    dialog.format_secondary_text(message)  # type: ignore
    dialog.show()  # type: ignore
    dialog.run()  # type: ignore
    dialog.destroy()  # type: ignore



from unittest import TestCase
from datetime import datetime, timezone
from eduvpn.storage import serialize_datetime, deserialize_datetime


utc = timezone.utc


class DateTimeSerializationTests(TestCase):
    def test_serialize_datetime(self):
        self.assertEqual(
            serialize_datetime(datetime(2022, 1, 3, 13, 46, 29, 0, utc)),
            '2022-01-03T13:46:29+00:00',
        )

    def test_deserialize_datetime(self):
        self.assertEqual(
            deserialize_datetime('2022-01-03T13:46:29'),
            datetime(2022, 1, 3, 13, 46, 29, 0, utc),
        )
        self.assertEqual(
            deserialize_datetime('2022-01-03T13:46:29.123456'),
            datetime(2022, 1, 3, 13, 46, 29, 123456, utc),
        )

        self.assertEqual(
            deserialize_datetime('2022-01-03T13:46:29+00:00'),
            datetime(2022, 1, 3, 13, 46, 29, 0, utc),
        )
        self.assertEqual(
            deserialize_datetime('2022-01-03T13:46:29-00:00'),
            datetime(2022, 1, 3, 13, 46, 29, 0, utc),
        )

    def test_serialize_roundtrip(self):
        dt = datetime.now(timezone.utc)
        self.assertEqual(
            deserialize_datetime(serialize_datetime(dt)),
            dt,
        )



from base64 import urlsafe_b64encode, b64encode, b64decode
import hashlib
import random
import logging
from datetime import timezone
from typing import Optional, List, NewType
from functools import lru_cache
from cryptography.x509.oid import NameOID
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from nacl.signing import VerifyKey
from nacl.exceptions import BadSignatureError
import nacl.bindings
import eduvpn
from eduvpn.settings import VERIFY_KEYS


logger = logging.getLogger(__name__)


def gen_code_challenge(code_verifier: str) -> bytes:
    """
    Transform the PKCE code verifier in a code challenge.

    args:
        code_verifier (str): a string generated with `gen_code_verifier()`
    """
    sha256 = hashlib.sha256(code_verifier.encode())
    encoded = urlsafe_b64encode(sha256.digest())
    return encoded.rstrip(b'=')


def gen_code_verifier(length: int = 128) -> str:
    """
    Generate a high entropy code verifier, used for PKCE.

    args:
        length (int): length of the code
    returns:
        str:
    """
    choices = 'abcdefghijklmnopqrstuvwxyz' \
              'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-._~'
    r = random.SystemRandom()
    return "".join(r.choice(choices) for _ in range(length))


def common_name_from_cert(pem_data: bytes) -> str:
    """
    Extract common name from client certificate.

    args:
        pem_data (str): PEM encoded certificate
    returns:
        str: the common name of the client certificate.
    """
    cert = x509.load_pem_x509_certificate(pem_data, default_backend())
    return cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value


def make_verifier(key: str) -> VerifyKey:
    """
    Create a NaCL verifier.

    args:
        key (str): A public key in minisign format
                   base64(<signature_algorithm> || <key_id> || <public_key>)
    returns:
        nacl.signing.VerifyKey: a nacl verifykey object
    """
    decoded = b64decode(key)[10:]
    return VerifyKey(decoded)


@lru_cache(1)
def make_verifiers() -> List[VerifyKey]:
    """
    Create a list of NaCL verifiers.

    returns:
        a list of nacl verify key objects.
    """
    # expecting format: base64(<signature_algorithm> || <key_id> || <public_key>)

    return [VerifyKey(b64decode(k)[10:]) for k in VERIFY_KEYS]


def verifier_to_str(verifier: VerifyKey) -> str:
    try:
        return str(verifier)
    except AttributeError:
        # Workaround for a bug in older versions of NaCL
        # https://github.com/pyca/pynacl/issues/205
        return str(verifier.__bytes__())


def validate(signature: str, content: bytes) -> bytes:
    decoded = b64decode(signature)[10:]
    verifiers = make_verifiers()

    logger.debug(f"Trying {len(verifiers)} verifiers")
    for f in verifiers:
        try:
            message = f.verify(smessage=content, signature=decoded)
            logger.debug(f"Used signature {verifier_to_str(f)}")
            return message
        except BadSignatureError:
            logger.debug(f"Skipping signature {verifier_to_str(f)}")
    raise BadSignatureError


def get_certificate_validity(certificate_text: str) -> Optional['eduvpn.session.Validity']:
    try:
        certificate_bytes = certificate_text.encode('ascii')
    except UnicodeEncodeError:
        logger.error(f"non-ascii certificate: {certificate_text!r}")
        return None
    certificate = x509.load_pem_x509_certificate(certificate_bytes, default_backend())
    from .session import Validity
    return Validity(
        start=certificate.not_valid_before.replace(tzinfo=timezone.utc),
        end=certificate.not_valid_after.replace(tzinfo=timezone.utc),
    )


SecretKey = NewType('SecretKey', str)
PulbicKey = NewType('PulbicKey', str)


class KeyPair:
    def __init__(self, public: PulbicKey, secret: SecretKey):
        self.public = public
        self.secret = secret


def generate_wireguard_keys() -> KeyPair:
    public_key, secret_key = nacl.bindings.crypto_box_keypair()
    public_key = b64encode(public_key).decode('ascii')
    secret_key = b64encode(secret_key).decode('ascii')
    return KeyPair(public_key, secret_key)



from typing import Optional
import logging
import webbrowser
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2.rfc6749.errors import (
    InvalidGrantError as InvalidOauthGrantError)
from .. import storage
from .. import nm
from .. import oauth2
from ..app import Application
from ..server import (
    AnyServer, ConfiguredServer, InstituteAccessServer,
    SecureInternetServer, OrganisationServer, CustomServer,
    SecureInternetLocation, Profile)
from .error import get_error_message
from ..session import active_session_states
from ..utils import run_in_background_thread


logger = logging.getLogger(__name__)


@run_in_background_thread('start-oauth-web-server')
def on_setup_oauth(app: Application, server: AnyServer):
    try:
        server_info = app.server_db.get_server_info(server)
    except Exception as e:
        logger.error("error getting server info", exc_info=True)
        enter_error_state_threadsafe(app, e)
        return

    def oauth_token_callback(oauth_session: Optional[OAuth2Session]):
        if oauth_session:
            app.interface_transition_threadsafe('oauth_setup_success', oauth_session)
        else:
            # The process has been canceled by the user through the app,
            # which already schedules the state transition so we don't need
            # to do that here.
            pass

    webserver, browser_url = oauth2.run_challenge_in_background(
        server_info.token_endpoint, server_info.authorization_endpoint, app.variant, oauth_token_callback)
    if isinstance(server, OrganisationServer):
        secure_internet = app.server_db.get_secure_internet_server(server.secure_internet_home)
        if secure_internet:
            browser_url = secure_internet.authentication_url(server, browser_url)
        else:
            logger.warning(f"missing 'secure internet server' for {server!r}")
    logger.info(f"opening browser with {browser_url}")
    app.interface_transition_threadsafe('ready_for_oauth_setup', webserver)
    webbrowser.open(browser_url)


@run_in_background_thread('oauth-refresh')
def on_refresh_oauth_token(app: Application,
                           server: AnyServer,
                           oauth_session: OAuth2Session):
    try:
        server_info = app.server_db.get_server_info(server)
    except Exception as e:
        logger.error("error getting server info", exc_info=True)
        enter_error_state_threadsafe(app, e)
        return
    try:
        oauth_session.refresh_token(token_url=server_info.token_endpoint)
    except InvalidOauthGrantError as e:
        logger.warning(f'Error refreshing OAuth token: {e}')
        app.interface_transition_threadsafe('oauth_refresh_failed')
    else:
        app.interface_transition_threadsafe('oauth_refresh_success')


@run_in_background_thread('load-server-info')
def on_start_connection(app: Application,
                        server: AnyServer,
                        oauth_session: OAuth2Session,
                        location: Optional[SecureInternetServer] = None):
    try:
        server_info = app.server_db.get_server_info(server)
    except Exception as e:
        logger.error("error getting server info", exc_info=True)
        enter_error_state_threadsafe(app, e)
        return
    profile_server_info = server_info
    profile_server: ConfiguredServer

    if isinstance(server, OrganisationServer):
        if not location:
            locations = [server for server in app.server_db.all()
                         if isinstance(server, SecureInternetServer)]
            app.interface_transition_threadsafe(
                'choose_secure_internet_location', server, oauth_session, locations)
            return
        else:
            try:
                location_info = app.server_db.get_server_info(location)
            except Exception as e:
                logger.error("error getting server info", exc_info=True)
                enter_error_state_threadsafe(app, e)
                return
            profile_server_info = location_info
            profile_server = SecureInternetLocation(server, location)
    else:
        profile_server = server

    all_profiles = profile_server_info.list_profiles(oauth_session)
    # Only show profiles with a supported protocol
    supported_profiles = [
        profile for profile in all_profiles
        if profile.has_supported_protocol and profile.supports_config(app.config)
    ]
    if not supported_profiles:
        message = "no profile with suppored protocol"
        logger.error(message, exc_info=False)
        enter_error_state_threadsafe(app, Exception(message))
        return

    app.interface_transition_threadsafe(
        'choose_profile', profile_server, oauth_session, supported_profiles)


@run_in_background_thread('configure-connection')
def on_chosen_profile(app: Application,
                      server: AnyServer,
                      oauth_session: OAuth2Session,
                      profile: Profile):
    country_code: Optional[str]

    if app.network_state.has_transition('disconnect'):
        app.network_transition('disconnect')
        on_disconnect(app)

    if isinstance(server, SecureInternetLocation):
        try:
            server_info = app.server_db.get_server_info(server.server)
            location_info = app.server_db.get_server_info(server.location)
        except Exception as e:
            logger.error("error getting server info", exc_info=True)
            enter_error_state_threadsafe(app, e)
            return
        auth_url = server.server.oauth_login_url
        api_url = location_info.api_endpoint
        connect_server_info = location_info
        country_code = server.location.country_code
        con_type = storage.ConnectionType.SECURE
        display_name = str(server.server)
        support_contact = server.location.support_contact
    else:
        try:
            server_info = app.server_db.get_server_info(server)
        except Exception as e:
            logger.error("error getting server info", exc_info=True)
            enter_error_state_threadsafe(app, e)
            return
        api_url = server_info.api_endpoint
        connect_server_info = server_info
        auth_url = server.oauth_login_url
        country_code = None
        display_name = str(server)
        if isinstance(server, InstituteAccessServer):
            con_type = storage.ConnectionType.INSTITUTE
            support_contact = server.support_contact
        elif isinstance(server, CustomServer):
            con_type = storage.ConnectionType.OTHER
            support_contact = []
        else:
            raise TypeError(server)

    try:
        connection = connect_server_info.connect(app, profile, oauth_session)
    except Exception as e:
        logger.error("error connecting", exc_info=True)
        enter_error_state_threadsafe(app, e)
        return
    validity = connection.validity
    storage.set_metadata(
        auth_url, oauth_session.token, server_info.token_endpoint,
        server_info.authorization_endpoint, api_url, display_name,
        support_contact, profile.id, con_type, country_code,
        validity.start, validity.end, connection.protocol)
    storage.set_auth_url(auth_url)

    logger.debug(f"starting connection: {connection!r}")

    # Apply the users settings to the ovpn file.
    if app.config.force_tcp:
        try:
            connection.force_tcp()
        except Exception as e:
            enter_error_state_threadsafe(app, e)
            return

    def connected_callback(result):
        logger.info(f"Finished saving network manager config: {result}")
        app.session_transition('new_session', server, validity)
        app.interface_transition('finished_configuring_connection')
        app.current_network_uuid = storage.get_uuid()
        assert app.current_network_uuid is not None
        nm.set_default_gateway(profile.use_as_default_gateway)
        app.network_transition('start_new_connection', server)

    @app.make_func_threadsafe
    def connect():
        connection.connect(connected_callback)

    connect()


def on_disconnect(app: Application):
    if isinstance(app.session_state, active_session_states):
        server = app.session_state.server
    else:
        return
    oauth_session = storage.get_oauth_session()
    if not oauth_session:
        return

    @run_in_background_thread('stop-connection')
    def inner():
        server_info = app.server_db.get_server_info(server)
        server_info.disconnect(oauth_session)

    inner()


def enter_error_state(app: Application, error: Exception):
    message = get_error_message(error)
    from .transition import go_to_main_state
    app.interface_transition('encountered_exception', message, go_to_main_state)


def enter_error_state_threadsafe(app: Application, error: Exception):
    app.make_func_threadsafe(enter_error_state)(app, error)



import logging
from argparse import ArgumentParser, Namespace
from sys import argv
from typing import List
from logging import getLogger
from eduvpn.storage import set_auth_url, set_metadata, ConnectionType
from eduvpn.remote import get_info
from eduvpn import actions
from eduvpn import menu

from eduvpn.menu import store_configuration

_logger = getLogger(__file__)


def list_(args: Namespace):
    args.match = None
    menu.search(args)


def search(args):
    menu.search(args)


def refresh(_):
    actions.refresh()


def activate(_):
    actions.activate()


def deactivate(_):
    actions.deactivate()


def status(_):
    actions.status()


def enroll(auth_url, display_name, support_contact, secure_internets, interactive: bool):
    api_url, oauth, token_endpoint, auth_endpoint = actions.fetch_token(auth_url)

    country_code = None
    if secure_internets and interactive:
        choice = menu.secure_internet_choice(secure_internets)
        if choice:
            base_url, country_code = choice
            api_url, _, _ = get_info(base_url)

    if secure_internets:
        con_type = ConnectionType.SECURE
    else:
        con_type = ConnectionType.INSTITUTE

    _logger.info(f"using {api_url} as api_url")
    profile_id = actions.get_profile(oauth, api_url, interactive=interactive)
    config, private_key, certificate = actions.get_config_and_keycert(oauth, api_url, profile_id)
    set_metadata(auth_url, oauth.token, token_endpoint, auth_endpoint, api_url, display_name, support_contact,
                 profile_id, con_type, country_code)
    set_auth_url(auth_url)
    store_configuration(config, private_key, certificate, interactive=interactive)


def interactive(args):
    auth_url, display_name, support_contact, secure_internets = menu.interactive(args)
    enroll(auth_url, display_name, support_contact, secure_internets, interactive=True)


def configure(args):
    """
    Configure in a non-interactive way based on supplied arguments
    """
    auth_url, display_name, support_contact, secure_internets = menu.configure(args)
    enroll(auth_url, display_name, support_contact, secure_internets, interactive=False)


def parse_eduvpn(args: List[str]):
    parser = ArgumentParser(description='The eduVPN command line client')
    subparsers = parser.add_subparsers(title='subcommands',
                                       description='Valid subcommands',
                                       help='append -h to any subcommand to see potential additional parameters')

    interactive_parser = subparsers.add_parser('interactive')
    interactive_parser.add_argument('match', nargs='?')
    interactive_parser.set_defaults(func=interactive)

    search_parsers = subparsers.add_parser('search')
    search_parsers.add_argument('match')
    search_parsers.set_defaults(func=search)

    configure_parser = subparsers.add_parser('configure')
    configure_parser.add_argument('match')
    configure_parser.add_argument('secure_internet', nargs='?')
    configure_parser.set_defaults(func=configure)

    subparsers.add_parser('refresh').set_defaults(func=refresh)
    subparsers.add_parser('list').set_defaults(func=list_)
    subparsers.add_parser('activate').set_defaults(func=activate)
    subparsers.add_parser('deactivate').set_defaults(func=deactivate)
    subparsers.add_parser('status').set_defaults(func=status)

    parsed = parser.parse_args(args)
    if hasattr(parsed, 'func'):
        parsed.func(parsed)
    else:
        parser.print_help()


def letsconnect_start(args):
    auth_url = args.url
    enroll(auth_url, display_name=None, support_contact=None, secure_internets=None, interactive=False)


def parse_letsconnect(args: List[str]):
    parser = ArgumentParser(description="The Let's Connect! command line client")
    subparsers = parser.add_subparsers(title='subcommands',
                                       description='Valid subcommands',
                                       help='append -h to any subcommand to see potential additional parameters')

    configure_parser = subparsers.add_parser('configure')
    configure_parser.add_argument('url')
    configure_parser.set_defaults(func=letsconnect_start)

    subparsers.add_parser('refresh').set_defaults(func=refresh)
    subparsers.add_parser('activate').set_defaults(func=activate)
    subparsers.add_parser('deactivate').set_defaults(func=deactivate)
    subparsers.add_parser('status').set_defaults(func=status)

    parsed = parser.parse_args(args)
    if hasattr(parsed, 'func'):
        parsed.func(parsed)
    else:
        parser.print_help()


def eduvpn():
    logging.basicConfig(level=logging.INFO)
    parse_eduvpn(argv[1:])


def letsconnect():
    logging.basicConfig(level=logging.INFO)
    parse_letsconnect(argv[1:])



from typing import Union, Optional, Iterable, List, Dict
from enum import Enum
import logging
import os
from urllib.parse import quote_plus
from requests_oauthlib import OAuth2Session
import nacl.exceptions
import eduvpn
from eduvpn import remote
from eduvpn import storage
from eduvpn import crypto
from eduvpn.i18n import extract_translation, retrieve_country_name
from eduvpn.settings import SERVER_URI, ORGANISATION_URI, FLAG_PREFIX, MAX_HTTP_RETRIES, MAX_HTTP_TIMEOUT
from eduvpn.utils import custom_server_oauth_url, add_retry_adapter


logger = logging.getLogger(__name__)
TranslatedStr = Union[str, Dict[str, str]]


class Protocol(Enum):
    OPENVPN = 'openvpn'
    WIREGUARD = 'wireguard'

    @property
    def is_supported(self) -> bool:
        if self.value == 'wireguard':
            from .nm import is_wireguard_supported
            return is_wireguard_supported()
        else:
            return True

    def supports_config(self, config: 'eduvpn.config.Configuration') -> bool:
        if config.force_tcp:
            if self is Protocol.WIREGUARD:
                # WireGuard cannot be forced over TCP.
                return False
        return True

    @property
    def config_type(self) -> str:
        return protocol_connection_config_mime_type[self]

    @classmethod
    def get_by_config_type(cls, config_type: str) -> 'Protocol':
        return connection_config_mime_type_protocol[config_type]


protocol_connection_config_mime_type = {
    # NOTE: Each protocol must have a unique type
    #       because the server ultimately decides on
    #       the protocol and the client uses the type
    #       to determine which one was actually chosen.
    Protocol.OPENVPN: 'application/x-openvpn-profile',
    Protocol.WIREGUARD: 'application/x-wireguard-profile',
}

connection_config_mime_type_protocol = dict(
    (proto, config_type)
    for (config_type, proto)
    in protocol_connection_config_mime_type.items()
)


class InstituteAccessServer:
    """
    A record from: https://disco.eduvpn.org/v2/server_list.json
    where: server_type == "institute_access"
    """

    def __init__(self,
                 base_url: str,
                 display_name: TranslatedStr,
                 support_contact: List[str] = [],
                 keyword_list: Optional[Union[str, List[str]]] = None):
        self.base_url = base_url
        self.display_name = display_name
        self.support_contact = support_contact
        if keyword_list is not None:
            if isinstance(keyword_list, str):
                keyword_list = [keyword_list]
            elif not isinstance(keyword_list, list):
                raise TypeError(keyword_list)
        self.keyword_list = keyword_list

    def __str__(self):
        return extract_translation(self.display_name)

    def __repr__(self):
        return f"<InstituteAccessServer {str(self)!r}>"

    @property
    def search_texts(self):
        texts = [str(self)]
        if self.keyword_list:
            texts.extend(self.keyword_list)
        return texts

    @property
    def oauth_login_url(self):
        return self.base_url

    def authentication_url(self, oauth_url: str) -> str:
        return oauth_url


class SecureInternetServer:
    """
    A record from: https://disco.eduvpn.org/v2/server_list.json
    where: server_type == "secure_internet"
    """

    def __init__(self,
                 base_url: str,
                 public_key_list: List[str],
                 country_code: str,
                 support_contact: List[str] = [],
                 authentication_url_template: Optional[str] = None):
        self.base_url = base_url
        self.public_key_list = public_key_list
        self.support_contact = support_contact
        self.country_code = country_code
        self.authentication_url_template = authentication_url_template

    def __str__(self):
        return f"{self.country_name} @ {self.base_url}"

    def __repr__(self):
        return f"<SecureInternetServer {str(self)!r}>"

    @property
    def country_name(self) -> str:
        return retrieve_country_name(self.country_code)

    @property
    def flag_path(self) -> Optional[str]:
        path = f'{FLAG_PREFIX}{self.country_code}@1,5x.png'
        if os.path.exists(path):
            return path
        else:
            return None

    @property
    def oauth_login_url(self):
        return self.base_url

    def authentication_url(
        self,
        organisation: 'OrganisationServer',
        oauth_url: str,
    ) -> str:
        if self.authentication_url_template is None:
            return oauth_url
        return (self.authentication_url_template
                .replace('@ORG_ID@', quote_plus(organisation.org_id))
                .replace('@RETURN_TO@', quote_plus(oauth_url)))


class OrganisationServer:
    """
    A record from: https://disco.eduvpn.org/v2/organization_list.json
    """

    def __init__(self,
                 secure_internet_home: str,
                 display_name: TranslatedStr,
                 org_id: str,
                 keyword_list: Dict[str, str] = {}):
        self.secure_internet_home = secure_internet_home
        self.display_name = display_name
        self.org_id = org_id
        self.keyword_list = keyword_list

    def __str__(self):
        return extract_translation(self.display_name)

    def __repr__(self):
        return f"<OrganisationServer {str(self)!r}>"

    @property
    def keyword(self) -> Optional[str]:
        if self.keyword_list:
            return extract_translation(self.keyword_list)
        return None

    @property
    def search_texts(self):
        texts = [str(self)]
        if self.keyword:
            texts.append(self.keyword)
        return texts

    @property
    def oauth_login_url(self):
        return self.secure_internet_home


class CustomServer:
    """
    A server defined by the user.
    """

    def __init__(self, address: str):
        self.address = address

    def __str__(self):
        return self.address

    def __repr__(self):
        return f"<CustomServer {str(self)!r}>"

    @property
    def oauth_login_url(self):
        return custom_server_oauth_url(self.address)

    def authentication_url(self, oauth_url: str) -> str:
        return oauth_url


class ServerInfo:
    def __init__(self, api_endpoint, token_endpoint, authorization_endpoint):
        self.api_endpoint = api_endpoint
        self.token_endpoint = token_endpoint
        self.authorization_endpoint = authorization_endpoint

    def api_call_endpoint(self, call: str) -> str:
        return self.api_endpoint + '/' + call

    def list_profiles(self, session: OAuth2Session) -> Iterable['Profile']:
        # https://github.com/eduvpn/documentation/blob/v3/API.md#info
        response = session.get(self.api_call_endpoint('info'))
        remote.check_response(response)
        profiles = response.json()['info']['profile_list']
        return [Profile(**data) for data in profiles]

    def get_profile(self, session: OAuth2Session, id: str) -> Optional['Profile']:
        profiles = [profile for profile in self.list_profiles(session)
                    if profile.id == id]
        if len(profiles) == 1:
            return profiles[0]
        else:
            return None

    def connect(self, app, profile: 'Profile', session: OAuth2Session) -> 'eduvpn.connection.Connection':
        # https://github.com/eduvpn/documentation/blob/v3/API.md#connect
        accept_types = ', '.join(
            proto.config_type
            for proto in profile.vpn_proto_list
            if proto.is_supported and proto.supports_config(app.config)
        )
        data = dict(profile_id=profile.id)

        keypair = None
        if Protocol.WIREGUARD in profile.vpn_proto_list:
            keypair = crypto.generate_wireguard_keys()
            data['public_key'] = keypair.public

        session = add_retry_adapter(session, MAX_HTTP_RETRIES)
        endpoint = self.api_call_endpoint('connect')
        try:
            response = session.post(
                endpoint,
                data=data,
                headers={'Accept': accept_types},
                timeout=MAX_HTTP_TIMEOUT)
        except Exception as e:
            msg = f"Got exception {e} requesting {endpoint}"
            logger.debug(msg)
            raise
        remote.check_response(response)
        from .connection import Connection
        connection = Connection.parse(response)
        if keypair:
            connection.set_secret_key(keypair.secret)
        return connection

    def disconnect(self, session: OAuth2Session):
        # https://github.com/eduvpn/documentation/blob/v3/API.md#disconnect
        # We allow a timeout for 1 second here
        try:
            session.post(self.api_call_endpoint('disconnect'), timeout=1)
        except:  # NOQA: ignore
            # NOTE: No need to check the response as,
            #       according to the API specification,
            #       this is a "best-effort" call.
            pass


class Profile:
    def __init__(self,
                 profile_id: str,
                 display_name: TranslatedStr,
                 default_gateway: Optional[bool] = None,
                 vpn_proto_list: Iterable[str] = frozenset('openvpn'),
                 ):
        self.profile_id = profile_id
        self.display_name = display_name
        self.default_gateway = default_gateway
        self.vpn_proto_list = frozenset(map(Protocol, vpn_proto_list))

    @property
    def id(self):
        return self.profile_id

    def __str__(self):
        return extract_translation(self.display_name)

    def __repr__(self):
        return f"<Profile id={self.id!r} {str(self)!r}>"

    @property
    def has_supported_protocol(self) -> bool:
        return any(proto.is_supported for proto in self.vpn_proto_list)

    def supports_config(self, config: 'eduvpn.config.Configuration') -> bool:
        return any(proto.supports_config(config) for proto in self.vpn_proto_list)

    @property
    def use_as_default_gateway(self) -> bool:
        if self.default_gateway is None:
            return False
        else:
            return self.default_gateway


class SecureInternetLocation:
    def __init__(self,
                 server: OrganisationServer,
                 location: SecureInternetServer):
        self.server = server
        self.location = location

    def __str__(self):
        return self.location.country_name

    def __repr__(self):
        return f"<SecureInternetLocation {self.server!r} {self.location!r}>"

    @property
    def image_path(self) -> Optional[str]:
        return self.location.flag_path

    @property
    def support_contact(self) -> List[str]:
        return self.location.support_contact

    @property
    def oauth_login_url(self):
        assert self.server.oauth_login_url == self.location.oauth_login_url
        return self.server.oauth_login_url

    def authentication_url(self, oauth_url: str) -> str:
        return self.location.authentication_url(self.server, oauth_url)


# typing aliases
PredefinedServer = Union[
    InstituteAccessServer,
    OrganisationServer,
    CustomServer,
]
ConfiguredServer = Union[
    InstituteAccessServer,
    SecureInternetLocation,
    CustomServer,
]
AnyServer = Union[
    PredefinedServer,
    ConfiguredServer,
]


def is_search_match(server: PredefinedServer, query: str) -> bool:
    if hasattr(server, 'search_texts'):
        return any(query.lower() in search_text.lower()
                   for search_text
                   in server.search_texts)  # type: ignore
    else:
        return False


class ServerSignatureError(Exception):
    def __init__(self, uri: str):
        self.uri = uri


class ServerDatabase:
    def __init__(self):
        # TODO load the servers from a cache
        self.servers = []
        self.is_loaded = False

    def update(self):
        """
        Download the list of institute and secure internet servers,
        and update this database.

        This method must be thread-safe.
        """
        new_servers = []
        try:
            server_list = remote.list_servers(SERVER_URI)
        except nacl.exceptions.BadSignatureError:
            raise ServerSignatureError(SERVER_URI)
        try:
            organisation_list = remote.list_organisations(ORGANISATION_URI)
        except nacl.exceptions.BadSignatureError:
            raise ServerSignatureError(ORGANISATION_URI)
        for server_data in server_list:
            server_type = server_data.pop('server_type')
            if server_type == 'institute_access':
                server = InstituteAccessServer(**server_data)
                new_servers.append(server)
            elif server_type == 'secure_internet':
                server = SecureInternetServer(**server_data)
                new_servers.append(server)
            else:
                raise ValueError(server_type, server_data)
        for organisation_data in organisation_list:
            server = OrganisationServer(**organisation_data)
            new_servers.append(server)
        # Atomic update of server map.
        # TODO keep custom other servers
        self.servers = new_servers
        self.is_loaded = True

    def all_configured(self) -> Iterable[ConfiguredServer]:
        "Return all configured servers."
        for key, data in storage.get_all_metadatas().items():
            if data['con_type'] == storage.ConnectionType.INSTITUTE:
                yield InstituteAccessServer(
                    base_url=data['api_base_uri'],
                    display_name=data['display_name'],
                    support_contact=data['support_contact'],
                    keyword_list=None,  # TODO
                )
            elif data['con_type'] == storage.ConnectionType.SECURE:
                secure_internet = SecureInternetServer(
                    base_url=data['api_base_uri'],
                    public_key_list=[],  # TODO
                    country_code=data['country_id'],
                    support_contact=data['support_contact'],
                    authentication_url_template=None,  # TODO
                )
                organisation = OrganisationServer(
                    secure_internet_home=data['api_base_uri'],
                    display_name=data['display_name'],
                    org_id='',  # TODO
                    keyword_list={},
                )
                yield SecureInternetLocation(organisation, secure_internet)
            elif data['con_type'] == storage.ConnectionType.OTHER:
                yield CustomServer(data['api_base_uri'])
            else:
                raise ValueError(data)

    def get_single_configured(self) -> Optional[ConfiguredServer]:
        auth_url = storage.get_auth_url()
        if auth_url is not None:
            for server in self.all_configured():
                if server.oauth_login_url == auth_url:
                    return server
        return None

    def all(self) -> Iterable[PredefinedServer]:
        "Return all servers."
        return iter(self.servers)

    def get_secure_internet_server(self, base_url: str) -> Optional[SecureInternetServer]:
        for server in self.all():
            if isinstance(server, SecureInternetServer) and server.base_url == base_url:
                return server
        return None

    def search(self, query: str) -> Iterable[PredefinedServer]:
        "Return all servers that match the search query."
        if query:
            for server in self.all():
                if is_search_match(server, query):
                    yield server
        else:
            yield from self.all()

    def get_server_info(self, server) -> ServerInfo:
        base_uri = server.oauth_login_url
        if not base_uri.endswith('/'):
            base_uri += '/'
        uri = base_uri + '.well-known/vpn-user-portal'
        json_data = remote.request(uri)
        info = json_data['api']['http://eduvpn.org/api#3']
        return ServerInfo(**info)



from typing import Optional, List
from requests_oauthlib import OAuth2Session
from eduvpn.settings import CLIENT_ID as OAUTH_CLIENT_ID
from .. import storage
from ..app import Application
from ..server import AnyServer, PredefinedServer
from . import event
from . import state


def enter_custom_address(app: Application, address: str) -> state.InterfaceState:
    """
    Enter an address for a custom server.
    """
    return state.ConfigureCustomServer(address)


def go_to_main_state(app: Application) -> state.InterfaceState:
    """
    If any servers have been configured, show the main state to select one.
    Otherwise, allow the user to configure a new server.
    """
    if not app.variant.use_configured_servers:
        return state.ConfigurePredefinedServer()
    configured_servers = list(app.server_db.all_configured())
    if configured_servers:
        return state.MainState(servers=configured_servers)
    else:
        return state.ConfigurePredefinedServer()


def enter_search_query(app: Application, search_query: str) -> state.InterfaceState:
    """
    Enter a search query for a predefined server.
    """
    results: Optional[List[PredefinedServer]]
    if search_query:
        if app.server_db.is_loaded:
            results = list(app.server_db.search(search_query))
        else:
            return state.PendingConfigurePredefinedServer(search_query)
    else:
        results = None
    return state.ConfigurePredefinedServer(search_query, results)


def create_new_oauth_session(token: str, token_endpoint: str) -> OAuth2Session:
    return OAuth2Session(
        client_id=OAUTH_CLIENT_ID,
        token=token,
        auto_refresh_url=token_endpoint,
    )


def connect_to_server(app: Application,
                      server: AnyServer,
                      renew: bool = False) -> state.InterfaceState:
    oauth_login_url = server.oauth_login_url  # type: ignore
    if renew:
        metadata = None
    else:
        metadata = storage.get_current_metadata(oauth_login_url)
    if metadata:
        # We've already configured this server.
        token, token_endpoint, *_ = metadata
        oauth_session = create_new_oauth_session(token, token_endpoint)
        event.on_refresh_oauth_token(app, server, oauth_session)
        return state.OAuthRefreshToken(app, server, oauth_session)
    else:
        # This is a new server that we need to configure first.
        event.on_setup_oauth(app, server)
        return state.OAuthSetupPending(server)



import logging
import time
import uuid
import enum
from functools import lru_cache
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from typing import Any, Optional, Tuple, Callable
from ipaddress import ip_interface, ip_address
from socket import AF_INET, AF_INET6
from configparser import ConfigParser

from .config import Configuration
from .ovpn import Ovpn
from .storage import set_uuid, get_uuid, write_ovpn
from .utils import cache, run_in_background_thread
from .crypto import SecretKey

_logger = logging.getLogger(__name__)

try:
    import gi

    gi.require_version('NM', '1.0')
    from gi.repository import NM, GLib  # type: ignore
except (ImportError, ValueError):
    _logger.warning("Network Manager not available")
    NM = None

try:
    import dbus
except ImportError:
    dbus = None


@lru_cache()
def get_client() -> 'NM.Client':
    """
    Create a new client object. We put this here so other modules don't need to import NM
    """
    return NM.Client.new(None)


@lru_cache()
def get_mainloop():
    """
    Create a new client object. We put this here so other modules don't need to import Glib
    """
    return GLib.MainLoop()


def get_iface() -> Optional[str]:
    """
    Get the interface as a string for an openvpn or wireguard connection if there is one
    """
    client = get_client()
    uuid = get_uuid()
    connection = client.get_connection_by_uuid(uuid)
    if connection is None:
        return None
    type = connection.get_connection_type()
    if type == 'vpn':
        return get_iface_ovpn()
    elif type == 'wireguard':
        return get_iface_wg()
    return None


def get_iface_ovpn() -> Optional[str]:
    """
    Get the interface as a string if there is a master device for OpenVPN
    """
    client = get_client()
    active_connection = client.get_primary_connection()
    if not active_connection:
        return None

    master = active_connection.get_master()
    if not master:
        return None

    return master.get_ip_iface()


def get_iface_wg() -> Optional[str]:
    """
    Get the interface as a string for wireguard
    """
    client = get_client()
    uuid = get_uuid()
    connection = client.get_connection_by_uuid(uuid)
    if connection is None:
        return None
    return connection.get_interface_name()


def get_ipv4() -> Optional[str]:
    """
    Get the ipv4 address for an openvpn or wireguard connection as a string if there is one
    """
    client = get_client()
    uuid = get_uuid()
    connection = client.get_connection_by_uuid(uuid)
    if connection is None:
        return None
    type = connection.get_connection_type()
    if type == 'vpn':
        return get_ipv4_ovpn()
    elif type == 'wireguard':
        return get_ipv4_wg()
    return None


def get_ipv6() -> Optional[str]:
    """
    Get the ipv6 address for an openvpn or wireguard connection as a string if there is one
    """
    client = get_client()
    uuid = get_uuid()
    connection = client.get_connection_by_uuid(uuid)
    if connection is None:
        return None
    type = connection.get_connection_type()
    if type == 'vpn':
        return get_ipv6_ovpn()
    elif type == 'wireguard':
        return get_ipv6_wg()
    return None


def get_ipv4_ovpn() -> Optional[str]:
    """
    Get the ipv4 address for an openvpn connection as a string if there is one
    """
    client = get_client()
    active_connection = client.get_primary_connection()
    if not active_connection:
        return None

    ip4_config = active_connection.get_ip4_config()
    addresses = ip4_config.get_addresses()
    if not addresses:
        return None
    return addresses[0].get_address()


def get_ipv6_ovpn() -> Optional[str]:
    """
    Get the ipv6 address for an openvpn connection as a string if there is one
    """
    client = get_client()
    active_connection = client.get_primary_connection()
    if not active_connection:
        return None

    ip6_config = active_connection.get_ip6_config()
    addresses = ip6_config.get_addresses()
    if not addresses:
        return None
    return addresses[0].get_address()


def get_ipv4_wg() -> Optional[str]:
    """
    Get the ipv4 address for a wireguard connection as a string if there is one
    """
    uuid = get_uuid()
    if uuid is None:
        return None
    client = get_client()
    connection = client.get_connection_by_uuid(uuid)
    if not connection:
        return None
    ip4_config = connection.get_setting_ip4_config()
    if not ip4_config:
        return None
    if ip4_config.get_num_addresses() == 0:
        return None
    address = ip4_config.get_address(0)
    return address.get_address()


def get_ipv6_wg() -> Optional[str]:
    """
    Get the ipv6 address for a wireguard connection as a string if there is one
    """
    uuid = get_uuid()
    if uuid is None:
        return None
    client = get_client()
    connection = client.get_connection_by_uuid(uuid)
    if not connection:
        return None
    ip6_config = connection.get_setting_ip6_config()
    if not ip6_config:
        return None
    if ip6_config.get_num_addresses() == 0:
        return None
    address = ip6_config.get_address(0)
    return address.get_address()


def nm_available() -> bool:
    """
    check if Network Manager is available
    """
    if NM is None:
        return False
    try:
        get_client()
    except Exception:
        return False
    else:
        return True


def nm_managed() -> bool:
    """
    check if any device of the primary connection is managed by Network Manager
    """
    client = get_client()
    active_connection = client.get_primary_connection()
    if active_connection is None:
        return False

    master_devices = active_connection.get_devices()
    if master_devices is None:
        return False

    return any(d.get_managed() for d in master_devices)


def get_existing_configuration_uuid() -> Optional[str]:
    uuid = get_uuid()
    if uuid is None:
        return None
    client = get_client()
    connection = client.get_connection_by_uuid(uuid)
    if connection is None:
        return None
    else:
        return uuid


def nm_ovpn_import(target: Path) -> Optional['NM.Connection']:
    """
    Use the Network Manager VPN config importer to import an OpenVPN configuration file.
    """
    vpn_infos = [i for i in NM.VpnPluginInfo.list_load() if i.get_name() == 'openvpn']

    if len(vpn_infos) != 1:
        raise Exception(f"Expected one openvpn VPN plugins, got: {len(vpn_infos)}")

    conn = vpn_infos[0].load_editor_plugin().import_(str(target))
    conn.normalize()
    return conn


def import_ovpn_and_certificate(ovpn: Ovpn, private_key: str, certificate: str) -> 'NM.SimpleConnection':
    """
    Import the OVPN string into Network Manager.
    """
    target_parent = Path(mkdtemp())
    target = target_parent / "eduVPN.ovpn"
    write_ovpn(ovpn, private_key, certificate, target)
    connection = nm_ovpn_import(target)
    rmtree(target_parent)
    return connection


def import_ovpn(ovpn: Ovpn) -> 'NM.SimpleConnection':
    """
    Import the OVPN string into Network Manager.
    """
    target_parent = Path(mkdtemp())
    target = target_parent / "eduVPN.ovpn"
    _logger.info(f"Writing configuration to {target}")
    with open(target, mode='w+t') as f:
        ovpn.write(f)
    connection = nm_ovpn_import(target)
    rmtree(target_parent)
    return connection


def add_connection_callback(client, result, callback=None):
    new_con = client.add_connection_finish(result)
    set_uuid(uuid=new_con.get_uuid())
    _logger.info(f"Connection added for uuid: {get_uuid()}")
    if callback is not None:
        callback(new_con is not None)


def add_connection(client: 'NM.Client', connection: 'NM.Connection', callback=None):
    _logger.info("Adding new connection")
    client.add_connection_async(connection=connection, save_to_disk=True, callback=add_connection_callback,
                                user_data=callback)


def update_connection_callback(remote_connection, result, callback=None):
    res = remote_connection.commit_changes_finish(result)
    _logger.debug(f"Connection updated for uuid: {get_uuid()}, result: {res}, remote_con: {remote_connection}")
    if callback is not None:
        callback(result)


def update_connection(old_con: 'NM.Connection', new_con: 'NM.Connection', callback=None):
    """
    Update an existing Network Manager connection with the settings from another Network Manager connection
    """
    _logger.info("Updating existing connection with new configuration")

    # Don't attempt to overwrite the uuid,
    # but reuse the one from the previous connection.
    # Refer to issue #269.
    for setting in new_con.get_settings():
        if setting.get_name() == 'connection':
            setting.props.uuid = old_con.get_uuid()
    old_con.replace_settings_from_connection(new_con)
    old_con.commit_changes_async(save_to_disk=True, cancellable=None, callback=update_connection_callback,
                                 user_data=callback)


def set_connection(client, new_connection, callback):
    uuid = get_uuid()
    if uuid:
        old_con = client.get_connection_by_uuid(uuid)
        if old_con:
            update_connection(old_con, new_connection, callback)
            return
    add_connection(client=client, connection=new_connection, callback=callback)


def save_connection(client: 'NM.Client', ovpn: Ovpn, private_key, certificate, callback=None):
    _logger.info("writing configuration to Network Manager")
    new_con = import_ovpn_and_certificate(ovpn, private_key, certificate)
    set_connection(client, new_con, callback)


def save_connection_with_config(client: 'NM.Client',
                                config,
                                private_key,
                                certificate,
                                callback=None,
                                ):
    ovpn = Ovpn.parse(config)
    settings = Configuration.load()
    if settings.force_tcp:
        ovpn.force_tcp()
    return save_connection(client, ovpn, private_key, certificate, callback)


def start_openvpn_connection(ovpn: Ovpn, *, callback=None):
    client = get_client()
    _logger.info("writing ovpn configuration to Network Manager")
    new_con = import_ovpn(ovpn)
    set_connection(client, new_con, callback)


def start_wireguard_connection(
    config: ConfigParser,
    *,
    secret_key: SecretKey,
    callback=None,
):
    client = get_client()
    _logger.info("writing wireguard configuration to Network Manager")

    ipv4s = []
    ipv6s = []
    for ip in config['Interface']['Address'].split(','):
        addr = ip_interface(ip.strip())
        if addr.version == 4:
            ipv4s.append(NM.IPAddress(AF_INET, str(addr.ip), addr.network.prefixlen))
        elif addr.version == 6:
            ipv6s.append(NM.IPAddress(AF_INET6, str(addr.ip), addr.network.prefixlen))

    dns4 = []
    dns6 = []
    dns_hostnames = []
    for dns_entry in config['Interface']['DNS'].split(','):
        stripped_entry = dns_entry.strip()
        try:
            address = ip_address(stripped_entry)
        # The entry is not an ip but a hostname
        # They need to be added to dns search domains
        except ValueError:
            dns_hostnames.append(stripped_entry)
        else:
            if address.version == 4:
                dns4.append(str(address))
            elif address.version == 6:
                dns6.append(str(address))

    profile = NM.SimpleConnection.new()
    s_con = NM.SettingConnection.new()
    s_con.set_property(NM.DEVICE_AUTOCONNECT, False)
    s_con.set_property(NM.SETTING_CONNECTION_ID, "eduvpn-wireguard")
    s_con.set_property(NM.SETTING_CONNECTION_TYPE, "wireguard")
    s_con.set_property(NM.SETTING_CONNECTION_UUID, str(uuid.uuid4()))
    s_con.set_property(NM.SETTING_CONNECTION_INTERFACE_NAME, "EduVPN-WG")

    # https://lazka.github.io/pgi-docs/NM-1.0/classes/WireGuardPeer.html#NM.WireGuardPeer
    peer = NM.WireGuardPeer.new()
    peer.set_endpoint(config['Peer']['Endpoint'], allow_invalid=False)
    peer.set_public_key(config['Peer']['PublicKey'], accept_invalid=False)
    for ip in config['Peer']['AllowedIPs'].split(','):
        peer.append_allowed_ip(ip.strip(), accept_invalid=False)

    s_ip4 = NM.SettingIP4Config.new()
    s_ip6 = NM.SettingIP6Config.new()

    for i in dns4:
        s_ip4.add_dns(i)
    for i in dns6:
        s_ip6.add_dns(i)
    for i in dns_hostnames:
        s_ip4.add_dns_search(i)
        s_ip6.add_dns_search(i)

    s_ip4.set_property(NM.SETTING_IP_CONFIG_METHOD, "manual")
    s_ip6.set_property(NM.SETTING_IP_CONFIG_METHOD, "manual")

    for i in ipv4s:
        s_ip4.add_address(i)
    for i in ipv6s:
        s_ip6.add_address(i)

    # https://lazka.github.io/pgi-docs/NM-1.0/classes/SettingWireGuard.html
    w_con = NM.SettingWireGuard.new()
    w_con.append_peer(peer)
    w_con.set_property(NM.SETTING_WIREGUARD_PRIVATE_KEY, secret_key)

    profile.add_setting(s_ip4)
    profile.add_setting(s_ip6)
    profile.add_setting(s_con)
    profile.add_setting(w_con)

    set_connection(client, profile, callback)


def get_cert_key(client: 'NM.Client', uuid: str) -> Tuple[str, str]:
    try:
        connection = client.get_connection_by_uuid(uuid)
        cert_path = connection.get_setting_vpn().get_data_item('cert')
    except Exception:
        _logger.error(f"Can't fetch stored VPN connection with uuid {uuid}")
        raise IOError("Can't fetch eduVPN profile")

    key_path = connection.get_setting_vpn().get_data_item('key')
    cert = open(cert_path).read()
    key = open(key_path).read()
    return cert, key


def activate_connection(client: 'NM.Client', uuid: str, callback=None):
    con = client.get_connection_by_uuid(uuid)
    _logger.info(f"activate_connection uuid: {uuid} connection: {con}")
    if con is None:
        # Temporary workaround, connection is sometimes created too
        # late while according to the logging the connection is already
        # created. Need to find the correct event to sync on.
        time.sleep(.1)
        GLib.idle_add(lambda: activate_connection(client, uuid, callback))
        return

    def activate_connection_callback(a_client, res, callback=None):
        try:
            result = a_client.activate_connection_finish(res)
        except Exception as e:
            _logger.error(e)
        else:
            _logger.info(F"activate_connection_async result: {result}")
        finally:
            if callback:
                callback()

    client.activate_connection_async(connection=con, callback=activate_connection_callback, user_data=callback)


def deactivate_connection(client: 'NM.Client', uuid: str, callback=None):
    connection = client.get_connection_by_uuid(uuid)
    if connection is None:
        _logger.warning(f"no connection to deactivate of uuid {uuid}")
        return
    type = connection.get_connection_type()
    if type == 'vpn':
        deactivate_connection_vpn(client, uuid, callback)
    elif type == 'wireguard':
        deactivate_connection_wg(client, uuid, callback)
    else:
        _logger.warning(f"unexpected connection type {type} of {uuid}")


def deactivate_connection_vpn(client: 'NM.Client', uuid: str, callback=None):
    con = client.get_primary_connection()
    _logger.debug(f"deactivate_connection uuid: {uuid} connection: {con}")
    if con:
        active_uuid = con.get_uuid()

        if uuid == active_uuid:
            def on_deactivate_connection(a_client: 'NM.Client', res, callback=None):
                try:
                    result = a_client.deactivate_connection_finish(res)
                except Exception as e:
                    _logger.error(e)
                else:
                    _logger.info(F"deactivate_connection_async result: {result}")
                finally:
                    if callback:
                        callback()

            client.deactivate_connection_async(active=con, callback=on_deactivate_connection, user_data=callback)
    else:
        _logger.info("No active connection to deactivate")


def deactivate_connection_wg(client: 'NM.Client', uuid: str, callback=None):
    devices = [
        device for device in client.get_all_devices()
        if device.get_type_description() == 'wireguard' and uuid in
        {conn.get_uuid() for conn in device.get_available_connections()}
    ]
    if not devices:
        _logger.warning("No WireGuard device to disconnect")
        return

    assert len(devices) == 1
    device = devices[0]

    @run_in_background_thread('wg-disconnect')
    def do_disconnect(a_device: 'NM.DeviceWireGuard', callback=None):
        try:
            result = a_device.disconnect()
        except Exception as e:
            _logger.error(e)
        else:
            _logger.info(F"disconnect result: {result}")
        finally:
            if callback:
                callback()

    _logger.debug(f"disconnect uuid: {uuid}")
    do_disconnect(device, callback)


class ConnectionState(enum.Enum):
    CONNECTING = enum.auto()
    CONNECTED = enum.auto()
    DISCONNECTED = enum.auto()
    FAILED = enum.auto()
    UNKNOWN = enum.auto()

    @classmethod
    def from_vpn_state(cls, state: 'NM.VpnConnectionState'):
        if state in [NM.VpnConnectionState.CONNECT,
                     NM.VpnConnectionState.IP_CONFIG_GET,
                     NM.VpnConnectionState.PREPARE]:
            return cls.CONNECTING
        elif state is NM.VpnConnectionState.ACTIVATED:
            return cls.CONNECTED
        elif state is NM.VpnConnectionState.DISCONNECTED:
            return cls.DISCONNECTED
        elif state is NM.VpnConnectionState.FAILED:
            return cls.FAILED
        elif state is NM.VpnConnectionState.NEED_AUTH:
            return cls.FAILED
        elif state is NM.VpnConnectionState.UNKNOWN:
            return cls.UNKNOWN
        else:
            raise ValueError(state)

    @classmethod
    def from_active_state(cls, state: 'NM.ActiveConnectionState'):
        if state is NM.ActiveConnectionState.ACTIVATING:
            return cls.CONNECTING
        elif state is NM.ActiveConnectionState.ACTIVATED:
            return cls.CONNECTED
        elif state in [NM.ActiveConnectionState.DEACTIVATED,
                       NM.ActiveConnectionState.DEACTIVATING]:
            return cls.DISCONNECTED
        elif state is NM.ActiveConnectionState.UNKNOWN:
            return cls.UNKNOWN
        else:
            raise ValueError(state)


def connection_status(client: 'NM.Client') -> Tuple[Optional[str], Optional['NM.ActiveConnectionState']]:
    con = client.get_primary_connection()
    if type(con) != NM.VpnConnection:
        return None, None
    uuid = con.get_uuid()
    status = con.get_state()
    return uuid, status


def get_connection_state() -> ConnectionState:
    client = get_client()
    uuid = get_uuid()
    connections = [connection for connection
                   in client.get_active_connections()
                   if connection.get_uuid() == uuid]
    if len(connections) == 1:
        connection = connections[0]
    else:
        return ConnectionState.DISCONNECTED
    if isinstance(connection, NM.VpnConnection):
        return ConnectionState.from_vpn_state(connection.get_vpn_state())
    elif isinstance(connection, NM.ActiveConnection):
        return ConnectionState.from_active_state(connection.get_state())
    else:
        _logger.warning(f"connection of unknown type: {connection!r}")
        return ConnectionState.UNKNOWN


def get_vpn_status(client: 'NM.Client') -> Tuple['NM.VpnConnectionState', 'NM.VpnConnectionStateReason']:
    vpns = [a for a in NM.Client.new(None).get_active_connections() if type(a) == NM.VpnConnection]
    if len(vpns) > 1:
        _logger.warning("more than one VPN connection active")
        return NM.VpnConnectionState.UNKNOWN, NM.VpnConnectionStateReason.UNKNOWN
    elif len(vpns) == 0:
        return NM.VpnConnectionState.UNKNOWN, NM.VpnConnectionStateReason.UNKNOWN
    else:
        return vpns[0].get_state(), vpns[0].get_state_reason()


def set_default_gateway(enable: bool):
    "If True, make the VPN connection the default gateway."
    _logger.info(f"setting default gateway: {enable}")
    client = get_client()
    uuid = get_uuid()
    connection = client.get_connection_by_uuid(uuid)
    ipv4_setting = connection.get_setting_ip4_config()
    ipv6_setting = connection.get_setting_ip6_config()
    ipv4_setting.set_property('never-default', not enable)
    ipv6_setting.set_property('never-default', not enable)
    connection.commit_changes(
        save_to_disk=True,
        cancellable=None,
    )


@lru_cache(maxsize=1)
def get_dbus() -> Optional['dbus.SystemBus']:
    """
    Get the DBus system bus.

    None is returned on failure.
    """
    if dbus is None:
        logging.debug("DBus module could not be imported")
        return None
    try:
        from dbus.mainloop.glib import DBusGMainLoop
        DBusGMainLoop(set_as_default=True)
        bus = dbus.SystemBus(private=True)
    except Exception:
        logging.debug("Unable to access dbus", exc_info=True)
        return None
    else:
        return bus


def subscribe_to_status_changes(
    callback: Callable[[ConnectionState], Any],
) -> bool:
    """
    Subscribe to all network status changes via DBus.

    The callback argument is called with the connection state and reason
    whenever they change.

    False is returned on failure.
    """
    bus = get_dbus()
    if bus is None:
        return False

    def wrapped_callback_vpn(state_code: 'dbus.UInt32', reason_code: 'dbus.UInt32'):
        state = NM.VpnConnectionState(state_code)
        callback(ConnectionState.from_vpn_state(state))

    def wrapped_callback_wg(state_code: 'dbus.UInt32', reason_code: 'dbus.UInt32'):
        state = NM.ActiveConnectionState(state_code)
        callback(ConnectionState.from_active_state(state))

    bus.add_signal_receiver(
        handler_function=wrapped_callback_vpn,
        dbus_interface='org.freedesktop.NetworkManager.VPN.Connection',
        signal_name='VpnStateChanged',
    )
    bus.add_signal_receiver(
        handler_function=wrapped_callback_wg,
        dbus_interface='org.freedesktop.NetworkManager.Connection.Active',
        signal_name='StateChanged',
    )
    return True


def action_with_mainloop(action: Callable):
    _logger.info("calling action with CLI mainloop")
    main_loop = get_mainloop()

    def quit_loop(*args, **kwargs):
        _logger.info("Quiting main loop, thanks!")
        main_loop.quit()

    action(callback=quit_loop)
    main_loop.run()


def save_connection_with_mainloop(config, private_key, certificate):
    action_with_mainloop(
        action=lambda callback: save_connection_with_config(get_client(), config, private_key, certificate, callback))


def activate_connection_with_mainloop(uuid):
    action_with_mainloop(action=lambda callback: activate_connection(get_client(), uuid, callback))


def deactivate_connection_with_mainloop(uuid):
    action_with_mainloop(action=lambda callback: deactivate_connection(get_client(), uuid, callback))


@cache
def is_wireguard_supported() -> bool:
    return hasattr(NM, 'WireGuardPeer')



from unittest import TestCase
from unittest.mock import patch

from eduvpn.ui import stats
from pathlib import Path
from tempfile import TemporaryDirectory


MOCK_IFACE = "mock"


def write_temp_stats_file(directory: Path, filename: str, total_bytes: int):
    f = open(directory / filename, "w")
    f.write(str(total_bytes))
    f.close()


def try_open(path: Path):
    if not path.is_file():
        return None
    return open(path, "r")


class TestStats(TestCase):
    @patch("eduvpn.ui.stats.NetworkStats.iface", MOCK_IFACE)
    def test_download(self):
        with TemporaryDirectory() as tempdir:
            # Create test data in the wanted files
            # Use the tempdir so it is cleaned up later
            values = [0, 37, 6166746255814, -43]
            filenames = [
                "rx_bytes",
                "rx_bytes_increased",
                "rx_bytes_increased_big",
                "rx_bytes_decreased",
            ]
            for index, value in enumerate(values):
                write_temp_stats_file(Path(tempdir), filenames[index], value)

            # Create the class instance
            class_ = stats.NetworkStats()

            # Add a file that does not exist
            filenames += ["idonotexist"]

            # For the expected values we want the human readable string
            # The last two are special
            #   - 0 B because the value has decreased
            #   - default text because the file does not exist
            expected_values = ["0 B", "37 B", "5.61 TB", "0 B", class_.default_text]

            # Loop over the files,
            # patch it and check if the expected value holds
            class_ = stats.NetworkStats()
            for index, filename in enumerate(filenames):
                location = Path(tempdir) / filename
                file_ = try_open(location)
                with patch("eduvpn.ui.stats.NetworkStats.download_file", file_):
                    self.assertEqual(class_.download, expected_values[index])
                    class_.cleanup()

    @patch("eduvpn.ui.stats.NetworkStats.iface", MOCK_IFACE)
    def test_upload(self):
        with TemporaryDirectory() as tempdir:
            # Create test data in the wanted files
            # Use the tempdir so it is cleaned up later
            values = [0, 1024, 237823782378, -47]
            filenames = [
                "tx_bytes",
                "tx_bytes_increased",
                "tx_bytes_increased_big",
                "tx_bytes_decreased",
            ]
            for index, value in enumerate(values):
                write_temp_stats_file(Path(tempdir), filenames[index], value)

            # Create the class instance
            class_ = stats.NetworkStats()

            # Add a file that does not exist
            filenames += ["idonotexist"]

            # For the expected values we want the human readable string
            # The last two are special
            #   - 0 B because the value has decreased
            #   - default text because the file does not exist
            expected_values = [
                "0 B",
                "1.00 kB",
                "221.49 GB",
                "0 B",
                class_.default_text,
            ]

            # Loop over the files,
            # patch it and check if the expected value holds
            for index, filename in enumerate(filenames):
                location = Path(tempdir) / filename
                file_ = try_open(location)
                with patch("eduvpn.ui.stats.NetworkStats.upload_file", file_):
                    self.assertEqual(class_.upload, expected_values[index])
                    class_.cleanup()



from typing import Optional
import logging
from gettext import gettext as _
from .server import ServerDatabase, ServerSignatureError
from . import nm
from . import storage
from .variants import ApplicationVariant
from .state_machine import StateMachine, InvalidStateTransition
from .config import Configuration
from .utils import run_in_background_thread, run_periodically, run_delayed, cancel_at_context_end


logger = logging.getLogger(__name__)

CHECK_NETWORK_INTERVAL = 1  # seconds


class Application:
    def __init__(self, variant: ApplicationVariant, make_func_threadsafe):
        self.variant = variant
        self.make_func_threadsafe = make_func_threadsafe
        from .network import NetworkState, InitialNetworkState
        from .session import (
            SessionState, InitialSessionState,
            SessionActiveState, SessionPendingExpiryState,
        )
        from .interface.state import InterfaceState, InitialInterfaceState
        self.network_state_machine: StateMachine[NetworkState]
        self.session_state_machine: StateMachine[SessionState]
        self.interface_state_machine: StateMachine[InterfaceState]
        self.network_state_machine = StateMachine(InitialNetworkState())
        self.session_state_machine = StateMachine(InitialSessionState())
        self.interface_state_machine = StateMachine(InitialInterfaceState())
        self.server_db = ServerDatabase()
        self.current_network_uuid: Optional[str] = None
        self.session_state_machine.register_level_callback(
            SessionActiveState, self.context_session_active)
        self.session_state_machine.register_level_callback(
            SessionPendingExpiryState, self.context_session_pending_expiry)

    def initialize(self):
        self.initialize_network()
        if self.variant.use_predefined_servers:
            self.initialize_server_db()
        self.config = Configuration.load()

    def initialize_network(self):
        """
        Determine the current network state.
        """
        # Check if a previous network configuration exists.
        uuid = nm.get_existing_configuration_uuid()
        if uuid:
            self.current_network_uuid = uuid
            # Check what server corresponds to the configuration.
            server = self.server_db.get_single_configured()
            if server is None:
                # There is a network configuration,
                # but no record of what server corresponds to it.
                self.session_transition('no_previous_session_found')
            else:
                validity = storage.get_current_validity(server.oauth_login_url)
                self.session_transition('found_active_session', server, validity)
        else:
            self.session_transition('no_previous_session_found')

        def on_network_update_callback(state):
            network.on_state_update_callback(self, state)

        from . import network
        if not nm.subscribe_to_status_changes(on_network_update_callback):
            logger.warning(
                "unable to subscribe to network updates; "
                "the application may not reflect the current state"
            )

        def check_network_state():
            from .network import check_network_state
            check_network_state(self)

        run_periodically(
            check_network_state,
            CHECK_NETWORK_INTERVAL,
            'check-network',
        )

    @run_in_background_thread('init-server-db')
    def initialize_server_db(self):
        """
        Load the lists of organisations and servers.
        """
        try:
            self.server_db.update()
        except ServerSignatureError as e:
            self.interface_transition_threadsafe(
                'encountered_exception',
                _("Received a bad signature from server {server}").format(server=e.uri))
        else:
            self.interface_transition_threadsafe('server_db_finished_loading')

    @property
    def network_state(self):
        """
        Get the current state of the network.
        """
        return self.network_state_machine.state

    @property
    def session_state(self):
        """
        Get the current state of the session.
        """
        return self.session_state_machine.state

    @property
    def interface_state(self):
        """
        Get the current state of the interface.
        """
        return self.interface_state_machine.state

    def connect_state_transition_callbacks(self, obj):
        """
        Register all state transition callback methods decorated with
        `@transition_callback()` and `@transition_edge_callback()`
        of an object.
        """
        from .network import NetworkState
        self.network_state_machine.connect_object_callbacks(
            obj, NetworkState)
        from .session import SessionState
        self.session_state_machine.connect_object_callbacks(
            obj, SessionState)
        from .interface.state import InterfaceState
        self.interface_state_machine.connect_object_callbacks(
            obj, InterfaceState)

    def _base_transition(self, state_machine, state_machine_name, transition, *args, **kwargs):
        """
        Perform a transition on a state machine.
        """
        logger.debug(
            f'{state_machine_name} transitioning: '
            f'{state_machine.state} -> {transition}')
        try:
            state_machine.transition(
                transition, self, *args, **kwargs)
        except InvalidStateTransition:
            logger.error(
                f'invalid {state_machine_name} state transition: '
                f'{state_machine.state} -> {transition}')
        else:
            logger.debug(
                f'{state_machine_name} transitioned: '
                f'{transition} -> {state_machine.state}')

    def network_transition(self, transition, *args, **kwargs):
        """
        Perform a transition on the network state.
        """
        self._base_transition(
            self.network_state_machine,
            'network',
            transition,
            *args,
            **kwargs,
        )

    def session_transition(self, transition, *args, **kwargs):
        """
        Perform a transition on the network state.
        """
        self._base_transition(
            self.session_state_machine,
            'session',
            transition,
            *args,
            **kwargs,
        )

    def interface_transition(self, transition, *args, **kwargs):
        """
        Perform a transition on the interface state.
        """
        self._base_transition(
            self.interface_state_machine,
            'interface',
            transition,
            *args,
            **kwargs,
        )

    def network_transition_threadsafe(self, transition, *args, **kwargs):
        """
        Threadsafe version of `network_transition()`.
        """
        transit = self.make_func_threadsafe(self.network_transition)
        transit(transition, *args, **kwargs)

    def session_transition_threadsafe(self, transition, *args, **kwargs):
        """
        Threadsafe version of `session_transition()`.
        """
        transit = self.make_func_threadsafe(self.session_transition)
        transit(transition, *args, **kwargs)

    def interface_transition_threadsafe(self, transition, *args, **kwargs):
        """
        Threadsafe version of `network_transition()`.
        """
        transit = self.make_func_threadsafe(self.interface_transition)
        transit(transition, *args, **kwargs)

    def context_session_active(self, state):
        expiry_duration = state.validity.pending_remaining.total_seconds()

        def on_session_pending_expiry():
            self.session_transition_threadsafe('pending_expiry')

        return cancel_at_context_end(run_delayed(on_session_pending_expiry, expiry_duration))

    def context_session_pending_expiry(self, state):
        expiry_duration = state.validity.remaining.total_seconds()

        def on_session_expired():
            self.session_transition_threadsafe('has_expired')

        return cancel_at_context_end(run_delayed(on_session_expired, expiry_duration))



import sys
from typing import Optional, Callable
import threading
from datetime import datetime
from email.utils import parsedate_to_datetime
from functools import lru_cache, partial, wraps
from gettext import gettext
from logging import getLogger
from os import path, environ
from sys import prefix
from requests import Session
from requests.adapters import HTTPAdapter, Retry


logger = getLogger(__file__)


def get_logger(name_space: str):
    return getLogger(name_space)


def add_retry_adapter(session: Session, retries: int):
    adapter = HTTPAdapter(max_retries=Retry(total=retries))
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session


@lru_cache(maxsize=1)
def get_prefix() -> str:
    """
    Returns the Python prefix where eduVPN is installed

    returns:
        path to Python installation prefix
    """
    target = 'share/eduvpn/builder/mainwindow.ui'
    local = path.dirname(path.dirname(path.abspath(__file__)))
    options = [local, path.expanduser('~/.local'), '/usr/local', prefix]
    for option in options:
        logger.debug(f"looking for '{target}' in '{option}'")
        if path.isfile(path.join(option, target)):
            return option
    raise Exception("Can't find eduVPN installation")


def get_config_dir() -> str:
    return environ.get("XDG_CONFIG_HOME", "~/.config")


def custom_server_oauth_url(address):
    if not address.startswith(('http://', 'https://')):
        address = f'https://{address}'
    if not address.endswith('/'):
        address += '/'
    return address


def thread_helper(func: Callable, *, name: Optional[str] = None) -> threading.Thread:
    """
    Runs a function in a thread

    args:
        func (lambda): a function to run in the background
    """
    thread = threading.Thread(target=func, name=name)
    thread.daemon = True
    thread.start()
    return thread


def run_in_background_thread(name: Optional[str] = None):
    """
    Decorator for functions that must always run
    in a background thread.
    """
    def decorator(func):

        @wraps(func)
        def background_func(*args, **kwargs):
            thread_helper(partial(func, *args, **kwargs), name=name)

        return background_func

    return decorator


def run_in_main_gtk_thread(func):
    """
    Decorator for functions that must always run
    in the main GTK thread.
    """
    from gi.repository import GLib

    @wraps(func)
    def main_gtk_thread_func(*args, **kwargs):
        GLib.idle_add(partial(func, *args, **kwargs))

    return main_gtk_thread_func


def run_periodically(func: Callable[[], None],
                     interval: float,
                     name: Optional[str] = None,
                     ) -> Callable[[], None]:
    """
    Run a funtion periodically in a background thread.

    The given function is called every `interval` seconds,
    until either it returns False
    or until the returned cancel callback is called.
    """
    if name is None:
        name = 'run-periodically'
    event = threading.Event()

    @run_in_background_thread(name)
    def run_periodic_thread():
        while 1:
            if event.wait(interval):
                return
            elif func() is False:
                return

    run_periodic_thread()
    return event.set


def run_delayed(func: Callable[[], None],
                delay: float,
                name: Optional[str] = None,
                ) -> Callable[[], None]:
    """
    Run a function with a delay.

    The given function is called once when
    `delay` seconds have passed.
    If the delay is less then zero, the function is called as soon as possible.
    Call the returned callback to cancel calling the delayed function.
    """
    if name is None:
        name = 'run-delayed'
    event = threading.Event()

    @run_in_background_thread(name)
    def run_delayed_thread():
        if event.wait(delay):
            return
        else:
            func()

    run_delayed_thread()
    return event.set


def cancel_at_context_end(cancel_callback: Callable[[], None]):
    """
    This generator is intended to be used with
    `state_machine.transition_level_callback`
    to cancel a thread when a state is exited.
    """
    try:
        yield
    finally:
        cancel_callback()


if sys.version_info < (3, 9):
    # Backported from Python 3.10
    # https://github.com/python/cpython/blob/3.10/Lib/functools.py#L651
    def cache(func):
        from functools import lru_cache
        return lru_cache(maxsize=None)(func)
else:
    from functools import cache  # noqa: W0611


def parse_http_date_header(date: str) -> datetime:
    return parsedate_to_datetime(date)


parse_http_expires_header = parse_http_date_header


def get_human_readable_bytes(total_bytes: int) -> str:
    """
    Helper function to calculate the human readable bytes.
    E.g. B, kB, MB, GB, TB.
    """
    suffix = ""
    hr_bytes = float(total_bytes)
    for suffix in ["B", "kB", "MB", "GB", "TB"]:
        if hr_bytes < 1024.0:
            break
        if suffix != "TB":
            hr_bytes /= 1024.0

    if suffix == "B":
        return f"{int(hr_bytes)} {suffix}"
    return f"{hr_bytes:.2f} {suffix}"


def translated_property(text):
    return property(lambda self: gettext(text))  # type: ignore



from unittest import mock, TestCase
from eduvpn.variants import EDUVPN
from eduvpn.oauth2.challenge import run_challenge, OAuth2Session, OAuthWebServer
from eduvpn.oauth2.http import stringify_image


class TestOauth(TestCase):

    @mock.patch.object(OAuth2Session, 'fetch_token')
    @mock.patch.object(OAuth2Session, 'authorization_url')
    @mock.patch.object(OAuthWebServer, 'run')
    @mock.patch('eduvpn.oauth2.challenge.webbrowser.open')
    def test_run_challenge(
            self,
            webbrowser_open: mock.MagicMock,
            webserver_run: mock.MagicMock,
            authorization_url: mock.MagicMock,
            fetch_token: mock.MagicMock,
    ):
        authorization_url.return_value = ["authorization_url", "state"]
        webserver_run.return_value = {'code': ["test"], 'state': ['state']}
        fetch_token.return_value = "token"

        run_challenge(authorization_endpoint="https://test", token_endpoint="https://test", app_variant=EDUVPN)

    def test_stringify_image(self):
        with self.assertRaises(FileNotFoundError):
            stringify_image("path")



from typing import (
    TypeVar, Generic, Any, Union, Optional, Callable,
    Type, Tuple, List, Dict, Set, Generator)
import enum
from contextlib import contextmanager


class TransitionEdge(enum.Enum):
    """
    The edge of a state lifetime.

    The edge is `enter` when the state starts,
    and `exit` when it ends.
    """

    enter = enum.auto()
    exit = enum.auto()


ENTER = TransitionEdge.enter
EXIT = TransitionEdge.exit

ANY_TRANSITION = object()
LEVEL_CONTEXT = object()


# typing aliases
State = TypeVar('State')
StateType = Type[State]
StateTargets = Union[StateType, Tuple[StateType, ...]]
TransitionCallback = Callable[[State, State], None]
LevelContext = Generator[None, None, None]
LevelCallback = Callable[[State], LevelContext]
TransitionCallbackRegistry = Dict[
    Optional[Tuple[StateType, TransitionEdge]],
    Set[TransitionCallback]]
LevelCallbackRegistry = Dict[StateType, Set[LevelCallback]]


TRANSITION_CALLBACK_MARKER = '__transition_callback_for_state'


def setattr_list_item(obj, attr, item):
    try:
        list_attr = getattr(obj, attr)
    except AttributeError:
        list_attr = []
        setattr(obj, attr, list_attr)
    list_attr.append(item)


def normalise_state_targets(state_targets: StateTargets) -> Tuple[StateType, ...]:
    "Normalise state targets argument to tuple."
    if not isinstance(state_targets, tuple):
        return (state_targets, )
    else:
        return state_targets


def transition_callback(state_targets: StateTargets):
    """
    Decorator factory to mark a method as a
    transition callback for all transitions.

    Note the argument is the base class for states of the state machine
    to register transition events of.
    Without this, there would be no way to know which
    state machine this callback targets.
    """
    state_targets = normalise_state_targets(state_targets)

    def decorator(func: TransitionCallback):
        for state_type in state_targets:
            setattr_list_item(func,
                              TRANSITION_CALLBACK_MARKER,
                              (ANY_TRANSITION, state_type))
        return func

    return decorator


def transition_edge_callback(edge: TransitionEdge,
                             state_targets: StateTargets):
    """
    Decorator factory to mark a method as a transition callback
    for specific state transition edges.
    """
    state_targets = normalise_state_targets(state_targets)

    def decorator(func: TransitionCallback):
        for state_type in state_targets:
            setattr_list_item(func,
                              TRANSITION_CALLBACK_MARKER,
                              (edge, state_type))
        return func

    return decorator


def transition_level_callback(state_targets: StateTargets):
    """
    Decorator factory to mark a method as a level callback
    for specific state contexts.
    """
    state_targets = normalise_state_targets(state_targets)

    def decorator(func: LevelCallback):
        for state_type in state_targets:
            setattr_list_item(func,
                              TRANSITION_CALLBACK_MARKER,
                              (LEVEL_CONTEXT, state_type))
        return func

    return decorator


def _find_transition_callbacks(obj: Any, base_state_type: Type[State]):
    for attr in dir(obj):
        try:
            callback = getattr(obj, attr)
        except Exception:
            # Gtk raises RuntimeError for some attributes.
            continue
        try:
            registrations = getattr(callback, TRANSITION_CALLBACK_MARKER)
        except AttributeError:
            pass
        else:
            for trigger, state_type in registrations:
                if issubclass(state_type, base_state_type):
                    yield callback, trigger, state_type


class InvalidStateTransition(Exception):
    def __init__(self, name: str):
        self.name = name


class StateMachine(Generic[State]):
    """
    State machine wrapper that allows registering transition callbacks.
    """

    def __init__(self, initial_state: State):
        self.initial_state = initial_state
        self._state = initial_state
        self._transition_callbacks: TransitionCallbackRegistry = {}
        self._level_callbacks: LevelCallbackRegistry = {}
        self._active_contexts: List[LevelContext] = []

    @property
    def state(self) -> State:
        """
        Obtain the current state.

        The state can be changed by calling `transition()`
        with the name of a transition of the current state.
        """
        # The state is behind a property to prevent setting it directly.
        return self._state

    def transition(self, transition: str, *args, **kwargs):
        """
        Transition to a new state.

        This method is *not* thread-safe,
        all calls should be made from the same thread.
        """
        old_state = self._state
        try:
            transition_func = getattr(old_state, transition)
        except AttributeError as e:
            raise InvalidStateTransition(transition) from e
        new_state = transition_func(*args, **kwargs)
        if new_state is not old_state:
            with self.trigger_callbacks(old_state, new_state):
                self._state = new_state
        return new_state

    @contextmanager
    def trigger_callbacks(self, old_state: State, new_state: State):
        self._exit_contexts()
        self._call_edge_callbacks(EXIT, old_state, new_state)
        try:
            yield
        finally:
            self._call_generic_callbacks(old_state, new_state)
            self._call_edge_callbacks(ENTER, old_state, new_state)
            self._enter_contexts(new_state)

    def trigger_initial_callbacks(self):
        with self.trigger_callbacks(self.initial_state, self.state):
            pass

    def register_generic_callback(self, callback: TransitionCallback):
        """
        Register a callback for all transitions.
        """
        self._transition_callbacks.setdefault(None, set()).add(callback)

    def register_edge_callback(self,
                               state_type: Type[State],
                               edge: TransitionEdge,
                               callback: TransitionCallback):
        """
        Register a callback for specific transition edges.
        """
        self._transition_callbacks.setdefault((state_type, edge), set()).add(callback)

    def register_level_callback(self,
                                state_type: Type[State],
                                context: LevelCallback):
        """
        Register a context function for specific state.

        The function must return a generator that yields once
        to capture the life of a state.
        similar to contextlib.contextmanager,
        The returned generator is executed up to the yield
        when the state is entered,
        Then, when the state is exited, the generator is resumed
        and is required to stop.
        """
        self._level_callbacks.setdefault(state_type, set()).add(context)

    def connect_object_callbacks(self, obj, base_state_type: Type[State]):
        """
        Register all state transition callback methods decorated with
        `@transition_callback()` and `@transition_edge_callback()`
        of an object.

        Provide the base class of states for this state machine
        as the second argument to filter registrations for this
        state machine only.
        Only method registered to a subclass of this base class
        will be connected.
        """
        iterator = _find_transition_callbacks(obj, base_state_type)
        for callback, trigger, state_type in iterator:
            if trigger is ANY_TRANSITION:
                # This callback targets all events.
                self.register_generic_callback(callback)
            elif trigger is LEVEL_CONTEXT:
                # This callback targets a state context.
                self.register_level_callback(state_type, callback)
            elif isinstance(trigger, TransitionEdge):
                # This callback targets a specific state edge.
                self.register_edge_callback(state_type, trigger, callback)
            else:
                raise TypeError(trigger)

    def _call_generic_callbacks(self, old_state: State, new_state: State):
        for callback in self._transition_callbacks.get(None, []):
            callback(old_state, new_state)

    def _call_edge_callbacks(self,
                             edge: TransitionEdge,
                             old_state: State,
                             new_state: State):
        state = new_state if edge is ENTER else old_state
        for callback in self._transition_callbacks.get((state.__class__, edge), []):
            callback(old_state, new_state)

    def _enter_contexts(self, state: State):
        for callback in self._level_callbacks.get(state.__class__, []):
            generator = callback(state)
            try:
                next(generator)
            except StopIteration:
                # The generator is expected to yield exactly once.
                raise RuntimeError("generator didn't yield")
            self._active_contexts.append(generator)

    def _exit_contexts(self):
        while self._active_contexts:
            generator = self._active_contexts.pop()
            try:
                next(generator)
            except StopIteration:
                # The generator is expected to stop after it has yielded once.
                pass
            else:
                raise RuntimeError("generator didn't stop")


class BaseState:
    """
    Base class for all state machine states.
    """

    def __repr__(self):
        fields = ','.join(f' {k}={v!r}' for k, v in self.__dict__.items())
        return f'<{self.__class__.__name__}{fields}>'

    def has_transition(self, name: str) -> bool:
        """
        Return True if this state defines the transition function.
        """
        return hasattr(self, name)

    def copy(self, **fields):
        """
        Return a copy of this state, with some fields altered.
        """
        return self.__class__(**{**self.__dict__, **fields})



from unittest import TestCase, skipIf

from eduvpn.nm import (nm_available, add_connection,
                       import_ovpn, get_client, get_mainloop)
from eduvpn.ovpn import Ovpn
from eduvpn.storage import get_uuid
from tests.mock_config import mock_config


@skipIf(not nm_available(), "Network manager not available")
class TestNm(TestCase):
    def test_nm_available(self):
        nm_available()

    def test_import_ovpn(self):
        ovpn = Ovpn.parse(mock_config)
        import_ovpn(ovpn)

    def test_get_mainloop(self):
        get_mainloop()

    def test_get_add_connection(self):
        client = get_client()
        ovpn = Ovpn.parse(mock_config)
        simple_connection = import_ovpn(ovpn)
        add_connection(client, simple_connection)

    def test_get_uuid(self):
        get_uuid()

    def test_activate_connection(self):
        get_client()
        get_uuid()
        # activate_connection(client, uuid)
        # get_cert_key(client, uuid)
        # connection_status(client, uuid)
        # deactivate_connection(client, uuid)
        # save_connection_with_config(client, ovpn, mock_key, mock_cert)



import gi
gi.require_version('Notify', '0.7')  # noqa: E402
from gi.repository import GdkPixbuf, Notify  # type: ignore

from .variants import ApplicationVariant


def initialize(app_variant: ApplicationVariant):
    Notify.init(app_variant.name)


class Notification:
    def __init__(self, app_variant: ApplicationVariant):
        self.app_variant = app_variant
        self.notification = None

    def _build(self):
        icon = GdkPixbuf.Pixbuf.new_from_file(self.app_variant.icon)  # type: ignore
        notification = Notify.Notification.new(self.app_variant.name)
        notification.set_icon_from_pixbuf(icon)
        notification.set_app_name(self.app_variant.name)
        return notification

    def show(self, title: str, message: str):
        if self.notification is None:
            self.notification = self._build()
        self.notification.update(title, message)  # type: ignore
        self.notification.show()  # type: ignore

    def hide(self):
        if self.notification is not None:
            self.notification.close()  # type: ignore
            self.notification = None



import threading


TIMEOUT = 5


def wait_for_state_change(state_machine) -> bool:
    """
    Wait for the next state change.

    Return False if the waiting timed out.
    """
    event = threading.Event()
    state_machine.register_generic_callback(lambda *a: event.set())
    return event.wait(TIMEOUT)


def wait_until_state(state_machine, state_type):
    """
    Wait for a specific state type.

    Return False if the waiting timed out.
    """
    while not isinstance(state_machine.state, state_type):
        if not wait_for_state_change(state_machine):
            return False
    return True


class StateTestCaseMixin:
    def assertReachesInterfaceState(self, app, state_type):
        if not wait_until_state(app.interface_state_machine, state_type):
            self.fail(
                f"timeout waiting for interface state {state_type!r},"
                f" current state is {app.interface_state!r}"
            )
        self.assertIsInstance(app.interface_state, state_type)

    def assertReachesNetworkState(self, app, state_type):
        if not wait_until_state(app.network_state_machine, state_type):
            self.fail(
                f"timeout waiting for network state {state_type!r},"
                f" current state is {app.network_state!r}"
            )
        self.assertIsInstance(app.network_state, state_type)



from unittest import TestCase
from eduvpn.cli import parse_eduvpn


class TestMain(TestCase):
    def test_parse_args(self):
        with self.assertRaises(SystemExit):
            parse_eduvpn(["test"])



from unittest import TestCase
from unittest.mock import patch, MagicMock
from eduvpn.actions import fetch_token, refresh, activate, deactivate
from oauthlib.oauth2.rfc6749.errors import InvalidGrantError


class TestCli(TestCase):
    @patch('eduvpn.actions.save_connection_with_mainloop')
    @patch('eduvpn.actions.create_keypair')
    @patch('eduvpn.actions.get_config')
    @patch('eduvpn.actions.list_profiles')
    @patch('eduvpn.actions.oauth2.run_challenge')
    @patch('eduvpn.actions.get_info')
    @patch('eduvpn.actions.get_client')
    def test_start(
            self,
            get_client: MagicMock,
            get_info: MagicMock,
            run_challenge: MagicMock,
            list_profiles: MagicMock,
            get_config: MagicMock,
            create_keypair: MagicMock,
            save_connection: MagicMock,
    ):
        create_keypair.return_value = ["cert", "key"]
        get_info.return_value = {"api_base_uri", "token_endpoint", "auth_endpoint"}
        list_profiles.return_value = [{'profile_id': 'internet'}]
        get_config.return_value = "a config file"

        args = MagicMock()
        args.match = "https://test"
        fetch_token(args)

    @patch('eduvpn.actions.OAuth2Session')
    @patch('eduvpn.actions.get_storage')
    @patch('eduvpn.actions.get_cert_key')
    @patch('eduvpn.actions.get_info')
    @patch('eduvpn.actions.check_certificate')
    @patch('eduvpn.actions.create_keypair')
    @patch('eduvpn.actions.get_config')
    @patch('eduvpn.actions.save_connection_with_mainloop')
    @patch('eduvpn.actions.get_client')
    def test_refresh(
            self,
            get_client: MagicMock,
            save_connection: MagicMock,
            get_config: MagicMock,
            create_keypair: MagicMock,
            check_certificate: MagicMock,
            get_info: MagicMock,
            get_cert_key: MagicMock,
            get_storage: MagicMock, _: MagicMock
    ):
        create_keypair.return_value = "key", "cert"
        check_certificate.return_value = False
        get_cert_key.return_value = "cert", "key"
        get_storage.return_value = "uuid", "auth_url", ({}, "", "", "", "", "", "", "", "", None, None, None)
        get_info.return_value = "api_base_uri", "token_endpoint", "auth_endpoint"
        refresh()

    @patch('eduvpn.actions.OAuth2Session')
    @patch('eduvpn.actions.get_storage')
    @patch('eduvpn.actions.get_cert_key')
    @patch('eduvpn.actions.get_info')
    @patch('eduvpn.actions.check_certificate')
    @patch('eduvpn.actions.create_keypair')
    @patch('eduvpn.actions.get_config')
    @patch('eduvpn.actions.save_connection_with_mainloop')
    @patch('eduvpn.actions.get_client')
    def test_refresh_invalid_signature(
            self,
            get_client: MagicMock,
            save_connection: MagicMock,
            get_config: MagicMock,
            create_keypair: MagicMock,
            check_certificate: MagicMock,
            get_info: MagicMock,
            get_cert_key: MagicMock,
            get_storage: MagicMock, oauth: MagicMock
    ):
        oauth.refresh_token = MagicMock(side_effect=InvalidGrantError("invalid signature"))

        create_keypair.return_value = "key", "cert"
        check_certificate.return_value = False
        get_cert_key.return_value = "cert", "key"
        get_storage.return_value = "uuid", "auth_url", ({}, "", "", "", "", "", "", "", "", None, None, None)
        get_info.return_value = "api_base_uri", "token_endpoint", "auth_endpoint"
        refresh()

    @patch('eduvpn.actions.refresh')
    @patch('eduvpn.actions.activate_connection_with_mainloop')
    @patch('eduvpn.actions.get_client')
    def test_activate(self, get_client, activate_connection, refresh):
        activate()

    @patch('eduvpn.actions.get_client')
    @patch('eduvpn.actions.deactivate_connection_with_mainloop')
    def test_deactivate(self, get_client, deactivate_connection_with_mainloop):
        deactivate()



import logging
from gettext import gettext as _

import gi
gi.require_version("Gtk", "3.0")  # noqa: E402
from gi.repository import GLib, Gio, Gtk

from .. import i18n
from .. import notify
from ..utils import run_in_main_gtk_thread
from ..state_machine import ENTER, transition_edge_callback
from ..variants import ApplicationVariant
from ..app import Application
from .. import network as network_state
from .. import session as session_state
from .ui import EduVpnGtkWindow


logger = logging.getLogger(__name__)


LOG_FORMAT = format_ = (
    '%(asctime)s - %(threadName)s - %(levelname)s - %(name)s'
    ' - %(filename)s:%(lineno)d - %(message)s'
)


NO_WINDOW_QUIT_NETWORK_STATES = (
    network_state.UnconnectedState,
    network_state.DisconnectedState,
)


class EduVpnGtkApplication(Gtk.Application):
    def __init__(self, *args, app_variant: ApplicationVariant, **kwargs):
        super().__init__(  # type: ignore
            *args,
            application_id=app_variant.app_id,
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,  # type: ignore
            **kwargs
        )

        self.app = Application(app_variant, run_in_main_gtk_thread)
        # Only allow a single window and track it on the app.
        self.window = None

        self.add_main_option(  # type: ignore
            'version',
            ord('v'),
            GLib.OptionFlags.NONE,  # type: ignore
            GLib.OptionArg.NONE,  # type: ignore
            "print version and exit",
            None,
        )
        self.add_main_option(  # type: ignore
            'debug',
            ord('d'),
            GLib.OptionFlags.NONE,  # type: ignore
            GLib.OptionArg.NONE,  # type: ignore
            "enable debug logging",
            None,
        )

    def do_startup(self):
        logger.debug('startup')
        Gtk.Application.do_startup(self)
        i18n.initialize(self.app.variant)
        notify.initialize(self.app.variant)
        self.connection_notification = notify.Notification(self.app.variant)
        self.app.connect_state_transition_callbacks(self)

    def do_shutdown(self):
        logger.debug('shutdown')
        self.connection_notification.hide()
        Gtk.Application.do_shutdown(self)

    def do_activate(self):
        logger.debug('activate')
        if not self.window:
            self.window = EduVpnGtkWindow(application=self)
            self.window.initialize()
            self.window.present()
            self.app.initialize()
        else:
            self.window.on_reopen_window()

    def do_command_line(self, command_line):
        logger.debug(f'command line: {command_line}')
        options = command_line.get_options_dict()
        # unpack the commandline args into a dict
        options = options.end().unpack()

        if 'version' in options:
            from eduvpn import __version__
            print(f"eduVPN Linux client version {__version__}")
            return 0

        if 'debug' in options:
            log_level = logging.DEBUG
        else:
            log_level = logging.INFO
        logging.basicConfig(level=log_level, format=LOG_FORMAT)

        self.activate()
        return 0

    def on_quit(self, action=None, param=None):
        logger.debug('quit')
        self.quit()

    def on_window_closed(self):
        logger.debug('window closed')
        if isinstance(self.app.network_state, NO_WINDOW_QUIT_NETWORK_STATES):
            # Quit the app if no connection is active on window close.
            logger.debug("window closed while no active connection")
            self.on_quit()

    @transition_edge_callback(ENTER, network_state.ConnectingState)
    def enter_ConnectingState(self, old_state, new_state):
        self.connection_notification.show(
            title=_("Connecting"),
            message=_(
                "The connection is being established. "
                "This should only take a moment."
            ))

    @transition_edge_callback(ENTER, network_state.ConnectedState)
    def enter_ConnectedState(self, old_state, new_state):
        self.connection_notification.show(
            title=_("Connected"),
            message=_("You are now connected to your server."))

    @transition_edge_callback(ENTER, session_state.SessionPendingExpiryState)
    def enter_SessionPendingExpiryState(self, old_state, new_state):
        self.connection_notification.show(
            title=_("Connected"),
            message=_(
                "Your session is about to expire. "
                "Renew the session to remain connected."
            ))

    @transition_edge_callback(ENTER, network_state.ReconnectingState)
    def enter_ReconnectingState(self, old_state, new_state):
        self.connection_notification.show(
            title=_("Connecting"),
            message=_(
                "The connection is being established. "
                "This should only take a moment."
            ))

    @transition_edge_callback(ENTER, network_state.DisconnectedState)
    def enter_DisconnectedState(self, old_state, new_state):
        self.connection_notification.hide()

    @transition_edge_callback(ENTER, session_state.SessionExpiredState)
    def enter_SessionExpiredState(self, old_state, new_state):
        self.connection_notification.show(
            title=_("Session expired"),
            message=_("Your session has expired."))

    @transition_edge_callback(ENTER, network_state.ConnectionErrorState)
    def enter_ConnectionErrorState(self, old_state, new_state):
        message = _("An error occured")
        if new_state.error:
            message = f"{message}: {new_state.error}"
        self.connection_notification.show(
            title=_("Connection Error"),
            message=message)

    @transition_edge_callback(ENTER, NO_WINDOW_QUIT_NETWORK_STATES)
    def enter_NoActiveConnection(self, old_state, new_state):
        if not self.window.is_visible():
            # Quit the app if no window is open when the connection is deactivated.
            logger.debug("connection deactivated while window closed")
            self.on_quit()
